import asyncio
import importlib
import inspect
import textwrap
from collections import defaultdict
from pathlib import Path
from typing import Any, cast, get_args, get_origin, get_type_hints

import aiohttp_cors
import click
import trafaret as t
from aiohttp import web
from aiohttp.web_urldispatcher import AbstractResource, DynamicResource
from pydantic import BaseModel, TypeAdapter
from trafaret.lib import _empty

import ai.backend.common.validators as tx
from ai.backend.common.api_handlers import BodyParam, PathParam, QueryParam
from ai.backend.common.json import pretty_json_str
from ai.backend.manager import __version__
from ai.backend.manager.api import ManagerStatus
from ai.backend.manager.api.session import UndefChecker
from ai.backend.manager.api.utils import Undefined
from ai.backend.manager.models.vfolder import VFolderPermissionValidator
from ai.backend.manager.server import global_subapp_pkgs


class ParseError(Exception):
    pass


def get_path_parameters(resource: AbstractResource) -> list[dict[str, Any]]:
    params = []
    if isinstance(resource, DynamicResource):
        if groupindex := resource._pattern.groupindex:
            params = [
                {"name": param, "in": "path", "required": True, "schema": {"type": "string"}}
                for param in groupindex.keys()
            ]
    return params


def flatten_or(scheme: t.Trafaret) -> list[t.Trafaret]:
    left, right = scheme.trafarets  # type: ignore[attr-defined]
    items = []
    if isinstance(left, t.Or):
        items.extend(flatten_or(left))
    else:
        items.append(left)

    if isinstance(right, t.Or):
        items.extend(flatten_or(right))
    else:
        items.append(right)
    return items


def _traverse(scheme: t.Trafaret) -> dict[str, Any]:
    if isinstance(scheme, t.Or):
        trafarets = flatten_or(scheme)
        valid_trafarets = [x for x in trafarets if not isinstance(x, (t.Null, UndefChecker))]
        if len(valid_trafarets) >= 2:
            return {"anyOf": [_traverse(s) for s in valid_trafarets]}
        scheme = valid_trafarets[0]
    if isinstance(scheme, t.Any):
        return {"type": "string"}
    if isinstance(scheme, t.Bool):
        return {"type": "boolean"}
    if isinstance(scheme, t.Dict):
        items = parse_trafaret_definition(scheme)
        properties = {d["name"]: d["schema"] for d in items}
        required_keys: list[str] = [d["name"] for d in items if d["required"]]
        return {"type": "object", "properties": properties, "required": required_keys}
    if isinstance(scheme, t.Enum):
        enum_values = scheme.variants  # type: ignore[attr-defined]
        return {"type": "string", "enum": enum_values}
    if isinstance(scheme, tx.DelimiterSeperatedList):
        return {
            "type": "array",
            "items": {"type": "string"},
            "description": (
                "List of items separated by a delimiter (default: comma). "
                "Items may contain spaces, but not the delimiter."
            ),
        }
    if isinstance(scheme, t.Float):
        resp = {"type": "integer"}
        if gte := scheme.gte:  # type: ignore[attr-defined]
            resp["minimum"] = gte
        if lte := scheme.lte:  # type: ignore[attr-defined]
            resp["maximum"] = lte
        return resp
    if isinstance(scheme, t.Int):
        resp = {"type": "integer"}
        if gte := scheme.gte:  # type: ignore[attr-defined]
            resp["minimum"] = gte
        if lte := scheme.lte:  # type: ignore[attr-defined]
            resp["maximum"] = lte
        return resp
    if isinstance(scheme, t.List):
        array_items = _traverse(scheme.trafaret)  # type: ignore[attr-defined]
        return {"type": "array", "items": array_items}
    if isinstance(scheme, t.Mapping):
        key_type = scheme.key.__class__.__name__  # type: ignore[attr-defined]
        value_type = scheme.value.__class__.__name__  # type: ignore[attr-defined]
        return {
            "type": "object",
            "description": f"Mapping({key_type} => {value_type})",
            "additionalProperties": True,
        }
    if isinstance(scheme, t.Regexp):
        pattern = scheme.regexp.pattern  # type: ignore[attr-defined]
        return {"type": "string", "pattern": str(pattern)}
    if isinstance(scheme, t.String):
        return {"type": "string"}
    if isinstance(scheme, t.ToBool):
        return {"type": "boolean"}
    if isinstance(scheme, tx.BinarySize):
        return {"type": "string", "description": "Size in binary format (e.g. 2KB, 3M, 4GiB)"}
    if isinstance(scheme, tx.Enum):
        if scheme.use_name:
            values = [e.name for e in scheme.enum_cls]
        else:
            values = [e.value for e in scheme.enum_cls]
        return {"type": "string", "enum": values}
    if isinstance(scheme, tx.JSONString):
        return {"type": "string", "description": "JSON string"}
    if isinstance(scheme, tx.Path):
        return {"type": "string", "description": "POSIX path"}
    if isinstance(scheme, tx.Slug):
        return {
            "type": "string",
            "description": (
                "String composed of alpha-numeric characters with hyphen and underscores in the middle "
                "(space and dot may be additionally allowed depending on the used locations)"
            ),
        }
    if isinstance(scheme, tx.TimeDuration):
        return {
            "anyOf": [
                {
                    "type": "string",
                    "description": (
                        "Human-readable time duration representation (e.g. 2y, 3d, 4m, 5h, 6s, ...)"
                    ),
                },
                {"type": "int", "minimum": 0, "description": "Number of seconds"},
                {"type": "float", "minimum": 0, "description": "Number of seconds"},
            ]
        }
    if isinstance(scheme, tx.URL):
        return {"type": "string", "format": "uri"}
    if isinstance(scheme, tx.UUID):
        return {"type": "string", "format": "uuid"}
    if isinstance(scheme, VFolderPermissionValidator):
        return {"type": "string", "enum": ["ro", "rw", "wd"]}

    if scheme == t.Email:
        return {"type": "string", "format": "email"}
    if scheme == t.IPv4:
        return {"type": "string", "format": "ipv4"}
    if scheme == t.IPv6:
        return {"type": "string", "format": "ipv6"}
    if scheme == t.URL:
        return {"type": "string", "format": "uri"}

    raise ParseError(f"Failed to convert unknown value {scheme} to OpenAPI type")


def parse_trafaret_value(scheme: t.Trafaret) -> tuple[dict[str, Any], bool]:
    optional = (
        isinstance(scheme, t.Or)
        and len([
            x
            for x in scheme.trafarets  # type: ignore[attr-defined]
            if isinstance(x, (t.Null, UndefChecker))
        ])
        > 0
    )

    return _traverse(scheme), optional


def parse_trafaret_definition(root: t.Dict) -> list[dict[str, Any]]:
    resp = []
    for key in root.keys:  # type: ignore[attr-defined]
        names: list[str] = []
        if isinstance(key, tx.AliasedKey):
            names.extend(key.names)
        elif isinstance(key, t.Key):
            names.append(key.name)  # type: ignore[attr-defined]
        schema, optional = parse_trafaret_value(key.trafaret)
        if key.default and key.default != _empty:
            if inspect.isclass(key.default):
                try:
                    default_value = key.default()
                except Exception:
                    default_value = str(key.default)
            elif isinstance(key.default, Undefined):
                default_value = None
            else:
                default_value = key.default

            schema["default"] = default_value
        if hasattr(key, "__openapi_desc__"):
            schema["description"] = key.__openapi_desc__
        resp += [{"name": names[0], "schema": schema, "required": not optional}]
    return resp


def extract_api_handler_params(
    handler: Any,
) -> tuple[type[BaseModel] | None, type[BaseModel] | None, type[BaseModel] | None]:
    """
    Extract Pydantic models from @api_handler decorated functions.

    Uses get_type_hints() to resolve string annotations from PEP 563
    (`from __future__ import annotations`). inspect.signature() returns
    string annotations in that case, which don't work with get_origin()/get_args().

    Returns:
        tuple of (body_model, query_model, path_model) - any can be None if not present
    """
    body_model: type[BaseModel] | None = None
    query_model: type[BaseModel] | None = None
    path_model: type[BaseModel] | None = None

    try:
        type_hints = get_type_hints(handler)
    except (ValueError, TypeError, NameError):
        return None, None, None

    for param_name, annotation in type_hints.items():
        if param_name == "return":
            continue

        origin = get_origin(annotation)
        if origin is None:
            continue

        args = get_args(annotation)
        if not args:
            continue

        model = args[0]
        if not (isinstance(model, type) and issubclass(model, BaseModel)):
            continue

        if origin is BodyParam:
            body_model = model
        elif origin is QueryParam:
            query_model = model
        elif origin is PathParam:
            path_model = model
        # HeaderParam is typically not included in OpenAPI request body/params

    return body_model, query_model, path_model


def generate_openapi(subapps: list[web.Application], verbose: bool = False) -> dict[str, Any]:
    openapi: dict[str, Any] = {
        "openapi": "3.1.0",
        "info": {
            "title": "Backend.AI Manager API",
            "description": "Backend.AI Manager REST API specification",
            "version": __version__,
            "contact": {
                "name": "Lablup Inc.",
                "url": "https://docs.backend.ai",
                "email": "contect@lablup.com",
            },
        },
        "components": {
            "securitySchemes": {
                "TokenAuth": {
                    "type": "ApiKey",
                    "in": "header",
                    "name": "Authorization: BackendAI",
                    "description": (
                        "Check https://docs.backend.ai/en/latest/manager/common-api/auth.html for"
                        " more information"
                    ),
                },
            },
            "schemas": {},
        },
        "paths": defaultdict(dict),
    }
    operation_id_mapping: defaultdict[str, int] = defaultdict(lambda: 0)
    for app in subapps:
        prefix = app.get("prefix", "root")
        for route in app.router.routes():
            resource = route.resource
            if not resource:
                continue

            if "_root_app" not in app:
                path = "/" + ("" if prefix == "root" else prefix) + resource.canonical
            else:
                path = resource.canonical
            method = route.method

            if method == "OPTIONS":
                continue

            operation_id = f"{prefix}.{route.handler.__name__}"
            if verbose:
                print(f"parsing {operation_id}")
            operation_id_mapping[operation_id] += 1
            if (operation_id_count := operation_id_mapping[operation_id]) > 1:
                operation_id += f".{operation_id_count}"

            description = []
            if route.handler.__doc__:
                description.append(textwrap.dedent(route.handler.__doc__))

            route_def = {
                "operationId": operation_id,
                "tags": [prefix],
                "responses": {"200": {"description": "Successful response"}},
            }
            parameters = []
            parameters.extend(get_path_parameters(resource))
            if hasattr(route.handler, "_backend_attrs"):
                preconds = []
                handler_attrs = route.handler._backend_attrs
                if handler_attrs.get("auth_required"):
                    route_def["security"] = [{"TokenAuth": []}]
                if auth_scope := handler_attrs.get("auth_scope"):
                    preconds.append(f"{auth_scope.capitalize()} privilege required.")
                if manager_status := handler_attrs.get("required_server_statuses"):
                    manager_status = cast(frozenset[ManagerStatus], manager_status)
                    if len(manager_status) == 1:
                        preconds.append(
                            f"Manager status required: {list(manager_status)[0].value.upper()}"
                        )
                    else:
                        preconds.append(
                            "Manager status required: one of "
                            f"{', '.join([e.value.upper() for e in sorted(manager_status)])}"
                        )
                if preconds:
                    description.append("\n**Preconditions:**")
                    for item in preconds:
                        description.append(f"* {item}")
                    description.append("")
                if request_scheme := handler_attrs.get("request_scheme"):
                    if isinstance(request_scheme, t.Dict):
                        parsed_definition = parse_trafaret_definition(request_scheme)
                        if method == "GET" or method == "DELETE":
                            parameters.extend([{**d, "in": "query"} for d in parsed_definition])
                        else:
                            properties = {d["name"]: d["schema"] for d in parsed_definition}
                            required_keys: list[str] = [
                                d["name"] for d in parsed_definition if d["required"]
                            ]
                            raw_examples = handler_attrs.get("request_examples") or []
                            examples = {
                                f"{operation_id}_Example{i}": {"value": e}
                                for e, i in zip(
                                    raw_examples, range(1, len(raw_examples) + 1), strict=True
                                )
                            }
                            route_def["requestBody"] = {
                                "content": {
                                    "application/json": {
                                        "schema": {
                                            "type": "object",
                                            "properties": properties,
                                            "required": required_keys,
                                        },
                                        "examples": examples,
                                    }
                                }
                            }
                    elif issubclass(request_scheme, BaseModel):
                        schema_name = request_scheme.__name__
                        request_schema = request_scheme.model_json_schema(
                            ref_template="#/components/schemas/{model}"
                        )

                        if additional_definitions := request_schema.pop("$defs", None):
                            openapi["components"]["schemas"].update(additional_definitions)
                        openapi["components"]["schemas"][schema_name] = request_schema
                        route_def["requestBody"] = {
                            "content": {
                                "application/json": {
                                    "schema": {"$ref": f"#/components/schemas/{schema_name}"}
                                }
                            }
                        }
                    else:
                        raise RuntimeError(
                            f"{request_scheme} not considered as a valid request type"
                        )

            # Handle @api_handler decorated functions (BodyParam, QueryParam, PathParam)
            if "requestBody" not in route_def:
                body_model, query_model, path_model = extract_api_handler_params(route.handler)

                if body_model is not None:
                    schema_name = body_model.__name__
                    request_schema = body_model.model_json_schema(
                        ref_template="#/components/schemas/{model}"
                    )
                    if additional_definitions := request_schema.pop("$defs", None):
                        openapi["components"]["schemas"].update(additional_definitions)
                    openapi["components"]["schemas"][schema_name] = request_schema
                    route_def["requestBody"] = {
                        "content": {
                            "application/json": {
                                "schema": {"$ref": f"#/components/schemas/{schema_name}"}
                            }
                        }
                    }

                if query_model is not None:
                    query_schema = query_model.model_json_schema(
                        ref_template="#/components/schemas/{model}"
                    )
                    properties = query_schema.get("properties", {})
                    required_keys = query_schema.get("required", [])
                    for prop_name, prop_schema in properties.items():
                        parameters.append({
                            "name": prop_name,
                            "in": "query",
                            "required": prop_name in required_keys,
                            "schema": prop_schema,
                        })

                if path_model is not None:
                    path_schema = path_model.model_json_schema(
                        ref_template="#/components/schemas/{model}"
                    )
                    properties = path_schema.get("properties", {})
                    for prop_name, prop_schema in properties.items():
                        # Check if already added by get_path_parameters
                        if not any(p["name"] == prop_name for p in parameters):
                            parameters.append({
                                "name": prop_name,
                                "in": "path",
                                "required": True,
                                "schema": prop_schema,
                            })

                # TODO: Extract response model from @api_handler decorated functions.
                # Currently, @api_handler returns APIResponse which wraps BaseResponseModel,
                # but the actual response model type is not captured in the return type annotation.
                #
                # To enable OpenAPI response schema generation, APIResponse should be made Generic:
                #
                #   # In api_handlers.py:
                #   TResponseModel = TypeVar("TResponseModel", bound=BaseResponseModel)
                #
                #   @dataclass
                #   class APIResponse(Generic[TResponseModel]):
                #       _status_code: int
                #       _data: Optional[TResponseModel]
                #
                #   # In handler:
                #   async def handler(...) -> APIResponse[MyResponseModel]:
                #       return APIResponse.build(200, MyResponseModel(...))
                #
                # Then we can extract the response model via:
                #   ret_type = get_type_hints(handler).get("return")
                #   if get_origin(ret_type) is APIResponse:
                #       response_model = get_args(ret_type)[0]

            route_def["parameters"] = parameters
            route_def["description"] = "\n".join(description)

            # Extract response schema from handler's return type annotation.
            # Supported return types:
            #   - BaseModel subclass: async def handler(...) -> MyResponseModel
            #   - list[BaseModel]: async def handler(...) -> list[MyResponseModel]
            #
            # Note: @api_handler decorated functions return APIResponse (not BaseModel),
            # so they are not captured here. See TODO comment above for future improvement.
            try:
                type_hints = get_type_hints(route.handler)
            except (NameError, AttributeError):
                # Skip type hint extraction for handlers with unresolvable forward references
                # (e.g., GraphQLView subclasses)
                type_hints = {}
            if (
                (ret_type := type_hints.get("return"))
                # For generic types like list[Model], get the origin (list).
                # For non-generic types like Model, use the type itself.
                and (response_cls := getattr(ret_type, "__origin__", ret_type))
                and (issubclass(response_cls, BaseModel) or issubclass(response_cls, list))
            ):
                response_schema: dict[str, Any]
                if issubclass(response_cls, list):
                    # Handle list[SomeModel] return type
                    arg: type[BaseModel]
                    (arg,) = get_args(ret_type)
                    schema_name = f"{arg.__name__}_List"
                    response_schema = TypeAdapter(list[arg]).json_schema(  # type: ignore[valid-type]
                        ref_template="#/components/schemas/{model}"
                    )
                elif issubclass(response_cls, BaseModel):
                    # Handle direct BaseModel return type
                    schema_name = response_cls.__name__
                    response_schema = response_cls.model_json_schema(
                        ref_template="#/components/schemas/{model}"
                    )

                else:
                    raise RuntimeError(f"{response_cls} not considered as a valid response type")

                # Extract nested model definitions (e.g., referenced models) to components
                if additional_definitions := response_schema.pop("$defs", None):
                    openapi["components"]["schemas"].update(additional_definitions)
                openapi["components"]["schemas"][schema_name] = response_schema
                route_def["responses"] = {
                    "200": {
                        "description": "",
                        "content": {
                            "application/json": {
                                "schema": {"$ref": f"#/components/schemas/{schema_name}"}
                            }
                        },
                    }
                }
            openapi["paths"][path][method.lower()] = route_def
    return openapi


async def generate() -> dict[str, Any]:
    cors_options = {
        "*": aiohttp_cors.ResourceOptions(  # type: ignore[no-untyped-call]
            allow_credentials=False, expose_headers="*", allow_headers="*"
        ),
    }

    subapps: list[web.Application] = []
    for subapp in global_subapp_pkgs:
        pkg = importlib.import_module("ai.backend.manager.api" + subapp)
        app, _ = pkg.create_app(cors_options)
        subapps.append(app)
    return generate_openapi(subapps, verbose=True)


@click.command()
@click.option(
    "--output",
    "-o",
    default=None,
    type=click.Path(dir_okay=False, writable=True),
    help="Output file path (default: stdout)",
)
def main(output: Path | None) -> None:
    """
    Generates OpenAPI specification of Backend.AI API.
    """
    openapi = asyncio.run(generate())
    if output is None:
        print(pretty_json_str(openapi))
    else:
        with output.open(mode="w") as fw:
            fw.write(pretty_json_str(openapi))


if __name__ == "__main__":
    main()



from .row import (
    BlueGreenSpec,
    DeploymentPolicyData,
    DeploymentPolicyRow,
    RollingUpdateSpec,
)

__all__ = (
    "BlueGreenSpec",
    "DeploymentPolicyData",
    "DeploymentPolicyRow",
    "RollingUpdateSpec",
)



"""
Lightweight entry point wrapper for the manager start-server command.

This module provides a minimal CLI entry point that defers heavy imports
until the command is actually executed, improving CLI startup time.
"""

from __future__ import annotations

from pathlib import Path

import click


@click.command()
@click.option(
    "-f",
    "--config-path",
    "--config",
    type=Path,
    default=None,
    help="The config file path. (default: ./manager.toml and /etc/backend.ai/manager.toml)",
)
@click.option(
    "--debug",
    is_flag=True,
    help="A shortcut to set `--log-level=DEBUG`",
)
@click.option(
    "--log-level",
    type=click.Choice(
        ["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG", "TRACE", "NOTSET"],
        case_sensitive=False,
    ),
    default="NOTSET",
    help="Set the logging verbosity level",
)
@click.pass_context
def main(
    ctx: click.Context,
    config_path: Path | None,
    debug: bool,
    log_level: str,
) -> None:
    """
    Start the manager service as a foreground process.

    This is a thin wrapper that defers the heavy import of server module.
    """
    from ai.backend.common.metrics.multiprocess import setup_prometheus_multiprocess_dir
    from ai.backend.logging import LogLevel

    # Must be called before importing server module (which triggers prometheus_client import)
    setup_prometheus_multiprocess_dir("manager")

    from ai.backend.manager.server import main as server_main

    ctx.invoke(
        server_main,
        config_path=config_path,
        debug=debug,
        log_level=LogLevel[log_level.upper()],
    )



"""Add object_storage_meta table

Revision ID: 902837cf18b4
Revises: 32764dcd0fad
Create Date: 2025-08-25 05:41:10.014103

"""

import sqlalchemy as sa
from alembic import op

from ai.backend.manager.models.base import GUID

# revision identifiers, used by Alembic.
revision = "902837cf18b4"
down_revision = "32764dcd0fad"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "object_storage_meta",
        sa.Column("id", GUID(), server_default=sa.text("uuid_generate_v4()"), nullable=False),
        sa.Column("storage_id", GUID(), nullable=False),
        sa.Column("bucket", sa.String(), nullable=False),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_object_storage_meta")),
    )
    op.create_unique_constraint(
        "uq_storage_id_bucket", "object_storage_meta", ["storage_id", "bucket"]
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint("uq_storage_id_bucket", "object_storage_meta", type_="unique")
    op.drop_table("object_storage_meta")
    # ### end Alembic commands ###



from dataclasses import dataclass
from typing import override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.data.resource.types import KeyPairResourcePolicyData
from ai.backend.manager.models.resource_policy import KeyPairResourcePolicyRow
from ai.backend.manager.repositories.base.creator import Creator
from ai.backend.manager.services.keypair_resource_policy.actions.base import (
    KeypairResourcePolicyAction,
)


@dataclass
class CreateKeyPairResourcePolicyAction(KeypairResourcePolicyAction):
    creator: Creator[KeyPairResourcePolicyRow]

    @override
    def entity_id(self) -> str | None:
        return None

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.CREATE


@dataclass
class CreateKeyPairResourcePolicyActionResult(BaseActionResult):
    keypair_resource_policy: KeyPairResourcePolicyData

    @override
    def entity_id(self) -> str | None:
        return self.keypair_resource_policy.name



"""Query conditions and orders for access tokens."""

from __future__ import annotations

import uuid
from collections.abc import Collection
from datetime import datetime
from typing import cast

import sqlalchemy as sa

from ai.backend.manager.models.endpoint import EndpointTokenRow
from ai.backend.manager.repositories.base import QueryCondition, QueryOrder


class AccessTokenConditions:
    """Query conditions for access tokens."""

    @staticmethod
    def by_ids(token_ids: Collection[uuid.UUID]) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return EndpointTokenRow.id.in_(token_ids)

        return inner

    @staticmethod
    def by_endpoint_id(endpoint_id: uuid.UUID) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return EndpointTokenRow.endpoint == endpoint_id

        return inner

    # Token string conditions
    @staticmethod
    def by_token_equals(value: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return EndpointTokenRow.token == value

        return inner

    @staticmethod
    def by_token_contains(value: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return EndpointTokenRow.token.contains(value)

        return inner

    # valid_until datetime conditions
    @staticmethod
    def by_valid_until_before(dt: datetime) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return cast(sa.sql.expression.ColumnElement[bool], EndpointTokenRow.valid_until < dt)

        return inner

    @staticmethod
    def by_valid_until_after(dt: datetime) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return cast(sa.sql.expression.ColumnElement[bool], EndpointTokenRow.valid_until > dt)

        return inner

    @staticmethod
    def by_valid_until_equals(dt: datetime) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return cast(sa.sql.expression.ColumnElement[bool], EndpointTokenRow.valid_until == dt)

        return inner

    # created_at datetime conditions
    @staticmethod
    def by_created_at_before(dt: datetime) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return EndpointTokenRow.created_at < dt

        return inner

    @staticmethod
    def by_created_at_after(dt: datetime) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return EndpointTokenRow.created_at > dt

        return inner

    @staticmethod
    def by_created_at_equals(dt: datetime) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return EndpointTokenRow.created_at == dt

        return inner

    @staticmethod
    def by_cursor_forward(cursor_id: str) -> QueryCondition:
        """Cursor condition for forward pagination (after cursor).

        Uses subquery to get created_at of the cursor row and compare.
        """
        cursor_uuid = uuid.UUID(cursor_id)

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            subquery = (
                sa.select(EndpointTokenRow.created_at)
                .where(EndpointTokenRow.id == cursor_uuid)
                .scalar_subquery()
            )
            return EndpointTokenRow.created_at < subquery

        return inner

    @staticmethod
    def by_cursor_backward(cursor_id: str) -> QueryCondition:
        """Cursor condition for backward pagination (before cursor).

        Uses subquery to get created_at of the cursor row and compare.
        """
        cursor_uuid = uuid.UUID(cursor_id)

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            subquery = (
                sa.select(EndpointTokenRow.created_at)
                .where(EndpointTokenRow.id == cursor_uuid)
                .scalar_subquery()
            )
            return EndpointTokenRow.created_at > subquery

        return inner


class AccessTokenOrders:
    """Query orders for access tokens."""

    @staticmethod
    def created_at(ascending: bool = True) -> QueryOrder:
        if ascending:
            return EndpointTokenRow.created_at.asc()
        return EndpointTokenRow.created_at.desc()



from dataclasses import dataclass
from typing import Self

from ai.backend.manager.repositories.keypair_resource_policy.repository import (
    KeypairResourcePolicyRepository,
)
from ai.backend.manager.repositories.types import RepositoryArgs


@dataclass
class KeypairResourcePolicyRepositories:
    repository: KeypairResourcePolicyRepository

    @classmethod
    def create(cls, args: RepositoryArgs) -> Self:
        repository = KeypairResourcePolicyRepository(args.db)

        return cls(
            repository=repository,
        )



from __future__ import annotations

from dataclasses import dataclass, field
from uuid import UUID

from ai.backend.common.data.model_deployment.types import DeploymentStrategy
from ai.backend.manager.data.deployment.types import (
    DeploymentMetadata,
    DeploymentNetworkSpec,
    ExecutionSpec,
    ImageIdentifierDraft,
    ModelRevisionSpec,
    ModelRevisionSpecDraft,
    MountInfo,
    ReplicaSpec,
    ResourceSpec,
)
from ai.backend.manager.data.image.types import ImageIdentifier
from ai.backend.manager.models.deployment_policy import BlueGreenSpec, RollingUpdateSpec


@dataclass
class VFolderMountsCreator:
    model_vfolder_id: UUID
    model_definition_path: str | None = None
    model_mount_destination: str = "/models"
    extra_mounts: list[MountInfo] = field(default_factory=list)


@dataclass
class ModelRevisionCreator:
    """Creator for model revision.

    Note: Uses image_id directly instead of image_identifier.
    The image_id is resolved by the GQL layer before being passed here.
    """

    image_id: UUID
    resource_spec: ResourceSpec
    mounts: VFolderMountsCreator
    execution: ExecutionSpec


@dataclass
class DeploymentCreator:
    metadata: DeploymentMetadata
    replica_spec: ReplicaSpec
    network: DeploymentNetworkSpec
    model_revision: ModelRevisionSpec
    policy: DeploymentPolicyConfig | None = None

    # Accessor properties for backward compatibility
    @property
    def image_identifier(self) -> ImageIdentifier:
        """Get the image identifier from model revision spec."""
        return self.model_revision.image_identifier

    @property
    def domain(self) -> str:
        """Get the domain name from metadata."""
        return self.metadata.domain

    @property
    def project(self) -> UUID:
        """Get the project ID from metadata."""
        return self.metadata.project

    @property
    def name(self) -> str:
        """Get the deployment name from metadata."""
        return self.metadata.name


@dataclass
class DeploymentCreationDraft:
    metadata: DeploymentMetadata
    replica_spec: ReplicaSpec
    network: DeploymentNetworkSpec
    draft_model_revision: ModelRevisionSpecDraft

    # Accessor properties for backward compatibility
    @property
    def image_identifier(self) -> ImageIdentifierDraft:
        """Get the requested image identifier from model revision spec."""
        return self.draft_model_revision.image_identifier

    @property
    def domain(self) -> str:
        """Get the domain name from metadata."""
        return self.metadata.domain

    @property
    def project(self) -> UUID:
        """Get the project ID from metadata."""
        return self.metadata.project

    @property
    def name(self) -> str:
        """Get the deployment name from metadata."""
        return self.metadata.name

    def to_creator(
        self,
        model_revision: ModelRevisionSpec,
    ) -> DeploymentCreator:
        return DeploymentCreator(
            metadata=self.metadata,
            replica_spec=self.replica_spec,
            network=self.network,
            model_revision=model_revision,
        )


@dataclass
class DeploymentPolicyConfig:
    """Configuration for deployment policy.

    Passed from GQL layer to service layer for policy creation/update.
    """

    strategy: DeploymentStrategy
    strategy_spec: RollingUpdateSpec | BlueGreenSpec
    rollback_on_failure: bool = False


@dataclass
class NewDeploymentCreator:
    metadata: DeploymentMetadata
    replica_spec: ReplicaSpec
    network: DeploymentNetworkSpec
    model_revision: ModelRevisionCreator
    policy: DeploymentPolicyConfig | None = None



"""Post-processor for invalidating kernel-related cache."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from ai.backend.common.types import AccessKey
from ai.backend.logging import BraceStyleAdapter

from .base import PostProcessor, PostProcessorContext

if TYPE_CHECKING:
    from ai.backend.manager.repositories.scheduler.repository import SchedulerRepository

log = BraceStyleAdapter(logging.getLogger(__name__))


class CacheInvalidationPostProcessor(PostProcessor):
    """Post-processor that invalidates kernel-related cache for affected access keys."""

    def __init__(self, repository: SchedulerRepository) -> None:
        self._repository = repository

    async def execute(self, context: PostProcessorContext) -> None:
        """Invalidate cache for all affected access keys."""
        affected_keys: set[AccessKey] = set()

        for info in context.result.successes:
            if info.access_key:
                affected_keys.add(info.access_key)
        for info in context.result.failures:
            if info.access_key:
                affected_keys.add(info.access_key)

        if affected_keys:
            await self._repository.invalidate_kernel_related_cache(list(affected_keys))
            log.debug(
                "Invalidated kernel-related cache for {} access keys",
                len(affected_keys),
            )



from __future__ import annotations

import uuid
from collections.abc import Collection

import sqlalchemy as sa

from ai.backend.manager.models.huggingface_registry import HuggingFaceRegistryRow
from ai.backend.manager.repositories.base import QueryCondition


class HuggingFaceRegistryConditions:
    """Query conditions for HuggingFace registries."""

    @staticmethod
    def by_ids(registry_ids: Collection[uuid.UUID]) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return HuggingFaceRegistryRow.id.in_(registry_ids)

        return inner



"""Types for user repository operations.

Contains Scope dataclasses for search operations.
"""

from __future__ import annotations

from collections.abc import Sequence
from dataclasses import dataclass
from uuid import UUID

import sqlalchemy as sa

from ai.backend.manager.errors.resource import DomainNotFound, ProjectNotFound
from ai.backend.manager.models.domain import DomainRow
from ai.backend.manager.models.group import AssocGroupUserRow, GroupRow
from ai.backend.manager.models.user import UserRow
from ai.backend.manager.repositories.base import ExistenceCheck, QueryCondition, SearchScope

__all__ = (
    "DomainUserSearchScope",
    "ProjectUserSearchScope",
)


@dataclass(frozen=True)
class DomainUserSearchScope(SearchScope):
    """Required scope for searching users within a domain.

    Used for domain_users query (domain admin+).
    """

    domain_name: str
    """Required. The domain to search within."""

    def to_condition(self) -> QueryCondition:
        """Convert scope to a query condition for UserRow."""
        domain_name = self.domain_name

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return UserRow.domain_name == domain_name

        return inner

    @property
    def existence_checks(self) -> Sequence[ExistenceCheck[str]]:
        """Return existence checks for scope validation."""
        return [
            ExistenceCheck(
                column=DomainRow.name,
                value=self.domain_name,
                error=DomainNotFound(self.domain_name),
            ),
        ]


@dataclass(frozen=True)
class ProjectUserSearchScope(SearchScope):
    """Required scope for searching users within a project.

    Used for project_users query (project member+).
    Requires JOIN with association_groups_users table.
    """

    project_id: UUID
    """Required. The project (group) to search within."""

    def to_condition(self) -> QueryCondition:
        """Convert scope to a query condition for AssocGroupUserRow."""
        project_id = self.project_id

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return AssocGroupUserRow.group_id == project_id

        return inner

    @property
    def existence_checks(self) -> Sequence[ExistenceCheck[UUID]]:
        """Return existence checks for scope validation."""
        return [
            ExistenceCheck(
                column=GroupRow.id,
                value=self.project_id,
                error=ProjectNotFound(str(self.project_id)),
            ),
        ]



"""add-resource-presets

Revision ID: 8e660aa31fe3
Revises: 01456c812164
Create Date: 2019-03-30 01:45:07.525096

"""

from decimal import Decimal
from typing import Any, cast

import sqlalchemy as sa
from alembic import op
from sqlalchemy.sql import text

from ai.backend.common.types import BinarySize, ResourceSlot
from ai.backend.manager.models.base import ResourceSlotColumn
from ai.backend.manager.models.resource_policy import keypair_resource_policies

# revision identifiers, used by Alembic.
revision = "8e660aa31fe3"
down_revision = "01456c812164"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "resource_presets",
        sa.Column("name", sa.String(length=256), nullable=False),
        sa.Column("resource_slots", ResourceSlotColumn(), nullable=False),
        sa.PrimaryKeyConstraint("name", name=op.f("pk_resource_presets")),
    )
    # Add initial fixtures for resource presets
    queries = """
    INSERT INTO resource_presets
    VALUES (
        'small',
        '{"cpu":"1","mem":"2147483648"}'::jsonb
    );
    INSERT INTO resource_presets
    VALUES (
        'small-gpu',
        '{"cpu":"1","mem":"2147483648","cuda.device":"1","cuda.shares":"0.5"}'::jsonb
    );
    INSERT INTO resource_presets
    VALUES (
        'medium',
        '{"cpu":"2","mem":"4294967296"}'::jsonb
    );
    INSERT INTO resource_presets
    VALUES (
        'medium-gpu',
        '{"cpu":"2","mem":"4294967296","cuda.device":"1","cuda.shares":"1.0"}'::jsonb
    );
    INSERT INTO resource_presets
    VALUES (
        'large',
        '{"cpu":"4","mem":"8589934592"}'::jsonb
    );
    INSERT INTO resource_presets
    VALUES (
        'large-gpu',
        '{"cpu":"4","mem":"8589934592","cuda.device":"2","cuda.shares":"2.0"}'::jsonb
    );
    """
    connection = op.get_bind()
    for query in queries.split(";"):
        if len(query.strip()) == 0:
            continue
        connection.execute(text(query))

    query = """
    SELECT name, total_resource_slots
    FROM keypair_resource_policies
    """
    result = connection.execute(text(query))
    updates = []
    for row in result:
        row_mapping = cast(dict[str, Any], row._mapping)
        converted = ResourceSlot(row_mapping["total_resource_slots"])
        if "mem" in converted:
            converted["mem"] = Decimal(BinarySize.from_str(converted["mem"]))
            updates.append((
                row_mapping["name"],
                converted,
            ))
    for name, slots in updates:
        update_query = (
            sa.update(keypair_resource_policies)
            .values(total_resource_slots=slots)
            .where(keypair_resource_policies.c.name == name)
        )
        connection.execute(update_query)

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("resource_presets")
    # ### end Alembic commands ###



"""Search resource group-scoped domains action and result types."""

from __future__ import annotations

from dataclasses import dataclass
from typing import override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.data.domain.types import DomainData
from ai.backend.manager.repositories.base import BatchQuerier
from ai.backend.manager.repositories.domain.types import DomainSearchScope
from ai.backend.manager.services.domain.actions.base import DomainAction


@dataclass
class SearchRGDomainsAction(DomainAction):
    """Action to search domains within a resource group scope.

    Returns only domains that are associated with the specified resource group
    through the sgroups_for_domains relationship.

    Args:
        scope: DomainSearchScope containing resource_group filter.
        querier: BatchQuerier containing additional filters, orders, and pagination.
    """

    scope: DomainSearchScope
    querier: BatchQuerier

    @override
    def entity_id(self) -> str | None:
        return self.scope.resource_group

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.SEARCH


@dataclass
class SearchRGDomainsActionResult(BaseActionResult):
    """Result of SearchRGDomainsAction.

    Args:
        items: List of domain data items in the current page.
        total_count: Total number of domains matching the filter within the scope.
        has_next_page: Whether there are more items after the current page.
        has_previous_page: Whether there are items before the current page.
    """

    items: list[DomainData]
    total_count: int
    has_next_page: bool
    has_previous_page: bool

    @override
    def entity_id(self) -> str | None:
        return None



"""Route-specific recorder for route coordinator."""

from __future__ import annotations

from uuid import UUID

from ai.backend.manager.sokovan.recorder.recorder import TransitionRecorder

# Route-specific type alias
RouteTransitionRecorder = TransitionRecorder[UUID]

__all__ = [
    "RouteTransitionRecorder",
]



from dataclasses import dataclass
from typing import Self

from ai.backend.common.data.permission.types import EntityType, FieldType, ScopeType

__all__ = [
    "EntityType",
    "FieldType",
    "ScopeType",
    "FieldRef",
    "ScopeId",
    "ObjectId",
]


@dataclass(frozen=True)
class FieldRef:
    """
    Represents a reference to a field within an entity.

    A field is a sub-resource that belongs to a parent entity.

    This class combines both the field type and its identifier to form
    a complete reference to a specific field instance.

    Attributes:
        field_type: The type of the field (FieldType enum).
        field_id: The unique identifier of the field.
    """

    field_type: FieldType
    field_id: str

    @classmethod
    def from_str(cls, val: str) -> Self:
        field_type, _, field_id = val.partition(":")
        return cls(field_type=FieldType(field_type), field_id=field_id)

    def to_str(self) -> str:
        return f"{self.field_type}:{self.field_id}"


@dataclass(frozen=True)
class ScopeId:
    scope_type: ScopeType
    scope_id: str

    @classmethod
    def from_str(cls, val: str) -> Self:
        scope_type, _, scope_id = val.partition(":")
        return cls(scope_type=ScopeType(scope_type), scope_id=scope_id)

    def to_str(self) -> str:
        return f"{self.scope_type}:{self.scope_id}"


@dataclass(frozen=True)
class ObjectId:
    entity_type: EntityType
    entity_id: str

    @classmethod
    def from_str(cls, val: str) -> Self:
        entity_type, _, entity_id = val.partition(":")
        return cls(entity_type=EntityType(entity_type), entity_id=entity_id)

    def to_str(self) -> str:
        return f"{self.entity_type}:{self.entity_id}"



from collections.abc import Sequence
from typing import override

from ai.backend.manager.sokovan.data import SessionWorkload, SystemSnapshot

from .sequencer import WorkloadSequencer


class FIFOSequencer(WorkloadSequencer):
    """
    A scheduling sequencer that implements First In, First Out (FIFO) sequencing.
    This sequencer will sequence the oldest workloads first.
    """

    @property
    @override
    def name(self) -> str:
        """
        Return the sequencer name for predicates.
        """
        return "FIFOSequencer"

    @override
    def success_message(self) -> str:
        """
        Return a message describing successful sequencing.
        """
        return "Sessions sequenced in first-in-first-out order"

    @override
    async def sequence(
        self,
        resource_group: str,
        system_snapshot: SystemSnapshot,
        workloads: Sequence[SessionWorkload],
    ) -> Sequence[SessionWorkload]:
        """
        Sequence the workloads in FIFO order.

        :param resource_group: The resource group (scaling group) name.
        :param system_snapshot: The current system snapshot containing resource state.
        :param workloads: A sequence of SessionWorkload objects to sequence.
        :return: A sequence of SessionWorkload objects in FIFO order.
        """
        return workloads  # Return the workloads in the order they were received



from __future__ import annotations

import uuid
from collections.abc import Collection

import sqlalchemy as sa

from ai.backend.manager.models.storage_namespace import StorageNamespaceRow
from ai.backend.manager.repositories.base import QueryCondition


class StorageNamespaceConditions:
    """Query conditions for storage namespaces."""

    @staticmethod
    def by_ids(namespace_ids: Collection[uuid.UUID]) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return StorageNamespaceRow.id.in_(namespace_ids)

        return inner



from dataclasses import dataclass
from typing import override

from ai.backend.common.types import AgentId, ImageID
from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.data.image.types import ImageData
from ai.backend.manager.services.image.actions.base import ImageAction
from ai.backend.manager.services.image.types import ImageRefData


@dataclass
class PurgeImageAction(ImageAction):
    image: ImageRefData
    agent_id: AgentId
    force: bool
    noprune: bool

    @override
    def entity_id(self) -> str | None:
        return None

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.PURGE


@dataclass
class PurgeImageActionResult(BaseActionResult):
    reserved_bytes: int
    purged_image: ImageData
    error: str | None

    @override
    def entity_id(self) -> str | None:
        return str(self.purged_image.id)


@dataclass
class PurgedImagesData:
    agent_id: AgentId
    purged_images: list[str]


@dataclass
class PurgeImagesKeyData:
    agent_id: AgentId
    images: list[ImageRefData]


# TODO: Remove this?
@dataclass
class PurgeImagesAction(ImageAction):
    keys: list[PurgeImagesKeyData]
    force: bool
    noprune: bool

    @override
    def entity_id(self) -> str | None:
        return None

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.PURGE


@dataclass
class PurgeImagesActionResult(BaseActionResult):
    total_reserved_bytes: int
    purged_images: list[PurgedImagesData]
    errors: list[str]

    @override
    def entity_id(self) -> str | None:
        return None


@dataclass
class PurgeImageByIdAction(ImageAction):
    image_id: ImageID

    @override
    def entity_id(self) -> str | None:
        return str(self.image_id)

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.PURGE


@dataclass
class PurgeImageByIdActionResult(BaseActionResult):
    image: ImageData

    @override
    def entity_id(self) -> str | None:
        return str(self.image.id)



"""Remove foreign key constraint from vfolders to users and projects

Revision ID: 59a622c31820
Revises: fdb2dcdb8811
Create Date: 2024-07-08 22:54:20.762521

"""

from alembic import op

# revision identifiers, used by Alembic.
revision = "59a622c31820"
down_revision = "fdb2dcdb8811"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.drop_constraint("fk_vfolders_user_users", "vfolders", type_="foreignkey")
    op.drop_constraint("fk_vfolders_group_groups", "vfolders", type_="foreignkey")
    op.drop_constraint("ck_vfolders_ownership_type_match_with_user_or_group", "vfolders")
    op.drop_constraint("ck_vfolders_either_one_of_user_or_group", "vfolders")


def downgrade() -> None:
    op.create_foreign_key("fk_vfolders_group_groups", "vfolders", "groups", ["group"], ["id"])
    op.create_foreign_key("fk_vfolders_user_users", "vfolders", "users", ["user"], ["uuid"])
    op.create_check_constraint(
        "ck_vfolders_ownership_type_match_with_user_or_group",
        "vfolders",
        "(ownership_type = 'user' AND \"user\" IS NOT NULL) OR "
        "(ownership_type = 'group' AND \"group\" IS NOT NULL)",
    )
    op.create_check_constraint(
        "ck_vfolders_either_one_of_user_or_group",
        "vfolders",
        '("user" IS NULL AND "group" IS NOT NULL) OR ("user" IS NOT NULL AND "group" IS NULL)',
    )



from __future__ import annotations

import enum
import uuid
from collections.abc import Mapping
from datetime import datetime
from typing import TYPE_CHECKING, Any, Final

import sqlalchemy as sa
from sqlalchemy.dialects import postgresql as pgsql
from sqlalchemy.exc import NoResultFound
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import Mapped, foreign, mapped_column, relationship, selectinload

from ai.backend.manager.models.base import (
    GUID,
    Base,
    SlugType,
)

if TYPE_CHECKING:
    from ai.backend.manager.models.domain import DomainRow
    from ai.backend.manager.models.group import GroupRow

__all__: Final[tuple[str, ...]] = (
    "NetworkRow",
    "NetworkType",
)


class NetworkType(enum.StrEnum):
    VOLATILE = "volatile"
    PERSISTENT = "persistent"
    HOST = "host"


def _get_project_join_condition() -> sa.ColumnElement[bool]:
    from ai.backend.manager.models.group import GroupRow

    return GroupRow.id == foreign(NetworkRow.project)


def _get_domain_join_condition() -> sa.ColumnElement[bool]:
    from ai.backend.manager.models.domain import DomainRow

    return DomainRow.name == foreign(NetworkRow.domain_name)


class NetworkRow(Base):  # type: ignore[misc]
    __tablename__ = "networks"

    id: Mapped[uuid.UUID] = mapped_column(
        "id", GUID, primary_key=True, server_default=sa.text("uuid_generate_v4()")
    )

    name: Mapped[str] = mapped_column("name", sa.String(length=128), nullable=False)
    """Name of the network"""
    ref_name: Mapped[str] = mapped_column("ref_name", sa.String(length=128), nullable=False)
    """Network ID generated by network plugin"""
    driver: Mapped[str] = mapped_column("driver", sa.String(length=64), nullable=False)
    """Network plugin in charge of the network created"""
    options: Mapped[Mapping[str, Any]] = mapped_column(
        "options", pgsql.JSONB, nullable=False, default="{}", server_default="{}"
    )

    project: Mapped[uuid.UUID] = mapped_column(
        "project",
        GUID,
        nullable=False,
    )
    domain_name: Mapped[str] = mapped_column(
        "domain_name",
        SlugType(length=64, allow_unicode=True, allow_dot=True),
        nullable=False,
    )

    created_at: Mapped[datetime | None] = mapped_column(
        "created_at",
        sa.DateTime(timezone=True),
        server_default=sa.text("now()"),
        nullable=True,
    )
    updated_at: Mapped[datetime | None] = mapped_column(
        "updated_at",
        sa.DateTime(timezone=True),
        server_default=sa.text("now()"),
        nullable=True,
    )

    project_row: Mapped[GroupRow] = relationship(
        "GroupRow",
        back_populates="networks",
        primaryjoin=_get_project_join_condition,
    )
    domain_row: Mapped[DomainRow] = relationship(
        "DomainRow",
        back_populates="networks",
        primaryjoin=_get_domain_join_condition,
    )

    def __init__(
        self,
        name: str,
        ref_name: str,
        driver: str,
        domain: str,
        project: uuid.UUID,
        *,
        options: Mapping[str, Any] | None = None,
    ) -> None:
        self.id = uuid.uuid4()
        self.name = name
        self.ref_name = ref_name
        self.driver = driver
        self.domain_name = domain
        self.project = project
        self.options = options or {}

    @classmethod
    async def get(
        cls,
        session: AsyncSession,
        network_id: uuid.UUID,
        load_project: bool = False,
        load_domain: bool = False,
    ) -> NetworkRow:
        query = sa.select(NetworkRow).filter(NetworkRow.id == network_id)
        if load_project:
            query = query.options(selectinload(cls.project_row))
        if load_domain:
            query = query.options(selectinload(cls.domain_row))
        row = await session.scalar(query)
        if not row:
            raise NoResultFound

        return row



"""Actions and results for Resource Usage Service."""

from __future__ import annotations

from dataclasses import dataclass
from typing import override

from ai.backend.common.data.permission.types import EntityType
from ai.backend.manager.actions.action import BaseAction, BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.repositories.base import (
    BatchQuerier,
    QueryCondition,
    QueryOrder,
    QueryPagination,
)
from ai.backend.manager.repositories.resource_usage_history import (
    DomainUsageBucketData,
    DomainUsageBucketSearchScope,
    ProjectUsageBucketData,
    ProjectUsageBucketSearchScope,
    UserUsageBucketData,
    UserUsageBucketSearchScope,
)

# Domain Usage Buckets


@dataclass
class DomainUsageBucketAction(BaseAction):
    """Base action for domain usage bucket operations."""

    @override
    @classmethod
    def entity_type(cls) -> EntityType:
        return EntityType.DOMAIN_USAGE_BUCKET


@dataclass
class SearchDomainUsageBucketsAction(DomainUsageBucketAction):
    """Action to search domain usage buckets."""

    pagination: QueryPagination
    conditions: list[QueryCondition]
    orders: list[QueryOrder]

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.SEARCH

    @override
    def entity_id(self) -> str | None:
        return None


@dataclass
class SearchDomainUsageBucketsActionResult(BaseActionResult):
    """Result of searching domain usage buckets."""

    items: list[DomainUsageBucketData]
    total_count: int

    @override
    def entity_id(self) -> str | None:
        return None


# Project Usage Buckets


@dataclass
class ProjectUsageBucketAction(BaseAction):
    """Base action for project usage bucket operations."""

    @override
    @classmethod
    def entity_type(cls) -> EntityType:
        return EntityType.PROJECT_USAGE_BUCKET


@dataclass
class SearchProjectUsageBucketsAction(ProjectUsageBucketAction):
    """Action to search project usage buckets."""

    pagination: QueryPagination
    conditions: list[QueryCondition]
    orders: list[QueryOrder]

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.SEARCH

    @override
    def entity_id(self) -> str | None:
        return None


@dataclass
class SearchProjectUsageBucketsActionResult(BaseActionResult):
    """Result of searching project usage buckets."""

    items: list[ProjectUsageBucketData]
    total_count: int

    @override
    def entity_id(self) -> str | None:
        return None


# User Usage Buckets


@dataclass
class UserUsageBucketAction(BaseAction):
    """Base action for user usage bucket operations."""

    @override
    @classmethod
    def entity_type(cls) -> EntityType:
        return EntityType.USER_USAGE_BUCKET


@dataclass
class SearchUserUsageBucketsAction(UserUsageBucketAction):
    """Action to search user usage buckets."""

    pagination: QueryPagination
    conditions: list[QueryCondition]
    orders: list[QueryOrder]

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.SEARCH

    @override
    def entity_id(self) -> str | None:
        return None


@dataclass
class SearchUserUsageBucketsActionResult(BaseActionResult):
    """Result of searching user usage buckets."""

    items: list[UserUsageBucketData]
    total_count: int

    @override
    def entity_id(self) -> str | None:
        return None


# Scoped Usage Bucket Actions


@dataclass
class SearchScopedDomainUsageBucketsAction(DomainUsageBucketAction):
    """Search domain usage buckets within scope."""

    scope: DomainUsageBucketSearchScope
    querier: BatchQuerier

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.SEARCH

    @override
    def entity_id(self) -> str | None:
        return f"{self.scope.resource_group}:{self.scope.domain_name}"


@dataclass
class SearchScopedDomainUsageBucketsActionResult(BaseActionResult):
    """Result of scoped domain usage bucket search."""

    items: list[DomainUsageBucketData]
    total_count: int

    @override
    def entity_id(self) -> str | None:
        return None


@dataclass
class SearchScopedProjectUsageBucketsAction(ProjectUsageBucketAction):
    """Search project usage buckets within scope."""

    scope: ProjectUsageBucketSearchScope
    querier: BatchQuerier

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.SEARCH

    @override
    def entity_id(self) -> str | None:
        return f"{self.scope.resource_group}:{self.scope.domain_name}:{self.scope.project_id}"


@dataclass
class SearchScopedProjectUsageBucketsActionResult(BaseActionResult):
    """Result of scoped project usage bucket search."""

    items: list[ProjectUsageBucketData]
    total_count: int

    @override
    def entity_id(self) -> str | None:
        return None


@dataclass
class SearchScopedUserUsageBucketsAction(UserUsageBucketAction):
    """Search user usage buckets within scope."""

    scope: UserUsageBucketSearchScope
    querier: BatchQuerier

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.SEARCH

    @override
    def entity_id(self) -> str | None:
        return f"{self.scope.resource_group}:{self.scope.domain_name}:{self.scope.project_id}:{self.scope.user_uuid}"


@dataclass
class SearchScopedUserUsageBucketsActionResult(BaseActionResult):
    """Result of scoped user usage bucket search."""

    items: list[UserUsageBucketData]
    total_count: int

    @override
    def entity_id(self) -> str | None:
        return None



"""Access token fetcher functions."""

from __future__ import annotations

from functools import lru_cache

from strawberry import Info
from strawberry.relay import PageInfo

from ai.backend.manager.api.gql.adapter import PaginationOptions, PaginationSpec
from ai.backend.manager.api.gql.base import encode_cursor
from ai.backend.manager.api.gql.deployment.types.access_token import (
    AccessToken,
    AccessTokenConnection,
    AccessTokenEdge,
    AccessTokenFilter,
    AccessTokenOrderBy,
)
from ai.backend.manager.api.gql.types import StrawberryGQLContext
from ai.backend.manager.models.endpoint.row import EndpointTokenRow
from ai.backend.manager.repositories.base import QueryCondition
from ai.backend.manager.repositories.deployment.options import (
    AccessTokenConditions,
    AccessTokenOrders,
)
from ai.backend.manager.services.deployment.actions.access_token.search_access_tokens import (
    SearchAccessTokensAction,
)


@lru_cache(maxsize=1)
def get_access_token_pagination_spec() -> PaginationSpec:
    """Get pagination specification for access tokens.

    Returns a cached PaginationSpec with:
    - Forward pagination: created_at DESC (newest first)
    - Backward pagination: created_at ASC
    """
    return PaginationSpec(
        forward_order=AccessTokenOrders.created_at(ascending=False),
        backward_order=AccessTokenOrders.created_at(ascending=True),
        forward_condition_factory=AccessTokenConditions.by_cursor_forward,
        backward_condition_factory=AccessTokenConditions.by_cursor_backward,
        tiebreaker_order=EndpointTokenRow.id.asc(),
    )


async def fetch_access_tokens(
    info: Info[StrawberryGQLContext],
    filter: AccessTokenFilter | None = None,
    order_by: list[AccessTokenOrderBy] | None = None,
    before: str | None = None,
    after: str | None = None,
    first: int | None = None,
    last: int | None = None,
    limit: int | None = None,
    offset: int | None = None,
    base_conditions: list[QueryCondition] | None = None,
) -> AccessTokenConnection:
    """Fetch access tokens with optional filtering, ordering, and pagination.

    Args:
        info: GraphQL context info
        filter: Optional filter criteria
        order_by: Optional ordering specification
        before/after/first/last: Cursor-based pagination parameters
        limit/offset: Offset-based pagination parameters
        base_conditions: Additional conditions to prepend (e.g., deployment_id filter)
    """
    processor = info.context.processors.deployment

    # Build querier using gql_adapter
    querier = info.context.gql_adapter.build_querier(
        PaginationOptions(
            first=first,
            after=after,
            last=last,
            before=before,
            limit=limit,
            offset=offset,
        ),
        get_access_token_pagination_spec(),
        filter=filter,
        order_by=order_by,
        base_conditions=base_conditions,
    )

    action_result = await processor.search_access_tokens.wait_for_complete(
        SearchAccessTokensAction(querier=querier)
    )

    nodes = [AccessToken.from_dataclass(data) for data in action_result.data]
    edges = [AccessTokenEdge(node=node, cursor=encode_cursor(str(node.id))) for node in nodes]

    return AccessTokenConnection(
        edges=edges,
        page_info=PageInfo(
            has_next_page=action_result.has_next_page,
            has_previous_page=action_result.has_previous_page,
            start_cursor=edges[0].cursor if edges else None,
            end_cursor=edges[-1].cursor if edges else None,
        ),
        count=action_result.total_count,
    )



"""Query conditions and orders for scheduler sessions, kernels, and users."""

from __future__ import annotations

from collections.abc import Collection
from typing import TYPE_CHECKING
from uuid import UUID

import sqlalchemy as sa

from ai.backend.manager.models.scaling_group.row import ScalingGroupRow

if TYPE_CHECKING:
    from ai.backend.manager.api.gql.base import UUIDEqualMatchSpec, UUIDInMatchSpec
    from ai.backend.manager.api.gql.kernel.types import KernelStatusInMatchSpec

from ai.backend.common.types import KernelId, SessionId
from ai.backend.manager.data.kernel.types import KernelStatus
from ai.backend.manager.data.session.types import KernelMatchType, SessionStatus
from ai.backend.manager.models.image import ImageRow
from ai.backend.manager.models.kernel import KernelRow
from ai.backend.manager.models.session import SessionRow
from ai.backend.manager.models.user import UserRow
from ai.backend.manager.repositories.base import QueryCondition, QueryOrder

# Default lookback period for fair share calculation (28 days)
DEFAULT_LOOKBACK_DAYS = 28


class SessionConditions:
    """Query conditions for sessions."""

    @staticmethod
    def by_ids(session_ids: Collection[SessionId]) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return SessionRow.id.in_(session_ids)

        return inner

    @staticmethod
    def by_statuses(statuses: Collection[SessionStatus]) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return SessionRow.status.in_(statuses)

        return inner

    @staticmethod
    def by_scaling_group(scaling_group: str) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return SessionRow.scaling_group_name == scaling_group

        return inner

    # Promotion handler conditions - optimized with EXISTS subqueries
    # These check kernel status conditions without loading kernel data

    @staticmethod
    def all_kernels_in_statuses(statuses: Collection[KernelStatus]) -> QueryCondition:
        """Filter sessions where ALL kernels are in the specified statuses.

        Uses NOT EXISTS to check that no kernel is outside the specified statuses.
        Also requires at least one kernel exists (sessions without kernels are excluded).
        """

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            # Subquery: check if any kernel is NOT in the specified statuses
            kernel_not_in_statuses = (
                sa.select(sa.literal(1))
                .select_from(KernelRow)
                .where(
                    KernelRow.session_id == SessionRow.id,
                    KernelRow.status.notin_(statuses),
                )
                .exists()
            )
            # Subquery: check if session has at least one kernel
            has_kernels = (
                sa.select(sa.literal(1))
                .select_from(KernelRow)
                .where(KernelRow.session_id == SessionRow.id)
                .exists()
            )
            # ALL: no kernel outside statuses AND has at least one kernel
            return sa.and_(~kernel_not_in_statuses, has_kernels)

        return inner

    @staticmethod
    def any_kernel_in_statuses(statuses: Collection[KernelStatus]) -> QueryCondition:
        """Filter sessions where at least one kernel is in the specified statuses.

        Uses EXISTS to check that at least one kernel is in the specified statuses.
        """

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            # Subquery: check if any kernel is in the specified statuses
            return (
                sa.select(sa.literal(1))
                .select_from(KernelRow)
                .where(
                    KernelRow.session_id == SessionRow.id,
                    KernelRow.status.in_(statuses),
                )
                .exists()
            )

        return inner

    @staticmethod
    def no_kernel_in_statuses(statuses: Collection[KernelStatus]) -> QueryCondition:
        """Filter sessions where no kernel is in the specified statuses.

        Uses NOT EXISTS to check that no kernel is in the specified statuses.
        Also requires at least one kernel exists (sessions without kernels are excluded).
        """

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            # Subquery: check if any kernel is in the specified statuses
            kernel_in_statuses = (
                sa.select(sa.literal(1))
                .select_from(KernelRow)
                .where(
                    KernelRow.session_id == SessionRow.id,
                    KernelRow.status.in_(statuses),
                )
                .exists()
            )
            # Subquery: check if session has at least one kernel
            has_kernels = (
                sa.select(sa.literal(1))
                .select_from(KernelRow)
                .where(KernelRow.session_id == SessionRow.id)
                .exists()
            )
            # NOT_ANY: no kernel in statuses AND has at least one kernel
            return sa.and_(~kernel_in_statuses, has_kernels)

        return inner

    @staticmethod
    def by_kernel_match(
        statuses: Collection[KernelStatus],
        match_type: KernelMatchType,
    ) -> QueryCondition:
        """Filter sessions by kernel status match type.

        Args:
            statuses: Kernel statuses to check against
            match_type: How to match kernel statuses (ALL/ANY/NOT_ANY)

        Returns:
            QueryCondition for the specified match type
        """
        match match_type:
            case KernelMatchType.ALL:
                return SessionConditions.all_kernels_in_statuses(statuses)
            case KernelMatchType.ANY:
                return SessionConditions.any_kernel_in_statuses(statuses)
            case KernelMatchType.NOT_ANY:
                return SessionConditions.no_kernel_in_statuses(statuses)


class SessionOrders:
    """Query orders for sessions."""

    @staticmethod
    def created_at(ascending: bool = True) -> QueryOrder:
        if ascending:
            return SessionRow.created_at.asc()
        return SessionRow.created_at.desc()

    @staticmethod
    def id(ascending: bool = True) -> QueryOrder:
        if ascending:
            return SessionRow.id.asc()
        return SessionRow.id.desc()


class KernelConditions:
    """Query conditions for kernels."""

    @staticmethod
    def by_id(kernel_id: KernelId) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return KernelRow.id == kernel_id

        return inner

    @staticmethod
    def by_ids(kernel_ids: Collection[KernelId]) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return KernelRow.id.in_(kernel_ids)

        return inner

    @staticmethod
    def by_id_filter_equals(spec: UUIDEqualMatchSpec) -> QueryCondition:
        """Factory for id equality filter."""

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.negated:
                return KernelRow.id != KernelId(spec.value)
            return KernelRow.id == KernelId(spec.value)

        return inner

    @staticmethod
    def by_id_filter_in(spec: UUIDInMatchSpec) -> QueryCondition:
        """Factory for id IN filter."""

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            kernel_ids = [KernelId(v) for v in spec.values]
            if spec.negated:
                return KernelRow.id.notin_(kernel_ids)
            return KernelRow.id.in_(kernel_ids)

        return inner

    @staticmethod
    def by_session_ids(session_ids: Collection[SessionId]) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return KernelRow.session_id.in_(session_ids)

        return inner

    @staticmethod
    def by_session_id_filter_equals(spec: UUIDEqualMatchSpec) -> QueryCondition:
        """Factory for session_id equality filter."""

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.negated:
                return KernelRow.session_id != SessionId(spec.value)
            return KernelRow.session_id == SessionId(spec.value)

        return inner

    @staticmethod
    def by_session_id_filter_in(spec: UUIDInMatchSpec) -> QueryCondition:
        """Factory for session_id IN filter."""

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            session_ids = [SessionId(v) for v in spec.values]
            if spec.negated:
                return KernelRow.session_id.notin_(session_ids)
            return KernelRow.session_id.in_(session_ids)

        return inner

    @staticmethod
    def by_statuses(statuses: Collection[KernelStatus]) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return KernelRow.status.in_(statuses)

        return inner

    @staticmethod
    def by_status_filter_in(spec: KernelStatusInMatchSpec) -> QueryCondition:
        """Factory for status IN filter."""

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if spec.negated:
                return KernelRow.status.notin_(spec.values)
            return KernelRow.status.in_(spec.values)

        return inner

    @staticmethod
    def by_scaling_group(scaling_group: str) -> QueryCondition:
        """Filter kernels by scaling group."""

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return KernelRow.scaling_group == scaling_group

        return inner

    @staticmethod
    def by_agent_id(agent_id: str) -> QueryCondition:
        """Filter kernels by agent ID."""

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return KernelRow.agent == agent_id

        return inner

    @staticmethod
    def by_cursor_forward(cursor_id: str) -> QueryCondition:
        """Cursor condition for forward pagination (after cursor).

        Uses subquery to get created_at of the cursor row and compare.
        """

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            subquery = (
                sa.select(KernelRow.created_at).where(KernelRow.id == cursor_id).scalar_subquery()
            )
            return KernelRow.created_at < subquery

        return inner

    @staticmethod
    def by_cursor_backward(cursor_id: str) -> QueryCondition:
        """Cursor condition for backward pagination (before cursor).

        Uses subquery to get created_at of the cursor row and compare.
        """

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            subquery = (
                sa.select(KernelRow.created_at).where(KernelRow.id == cursor_id).scalar_subquery()
            )
            return KernelRow.created_at > subquery

        return inner

    @staticmethod
    def for_fair_share_observation(
        scaling_group: str,
    ) -> QueryCondition:
        """Filter kernels that need fair share observation.

        Includes:
        1. Running kernels (terminated_at IS NULL) with starts_at set
        2. Recently terminated kernels with unobserved periods
           (terminated_at > last_observed_at, within lookback window)

        The lookback_days is fetched from scaling_groups.fair_share_spec via subquery.

        Args:
            scaling_group: The scaling group to filter

        Returns:
            QueryCondition for fair share observation targets
        """

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            # Subquery to get lookback_days from scaling_group's fair_share_spec
            # Falls back to DEFAULT_LOOKBACK_DAYS if not set
            lookback_days_subquery = (
                sa.select(
                    sa.func.coalesce(
                        sa.cast(
                            ScalingGroupRow.fair_share_spec["lookback_days"].as_string(),
                            sa.Integer,
                        ),
                        DEFAULT_LOOKBACK_DAYS,
                    )
                )
                .where(ScalingGroupRow.name == scaling_group)
                .scalar_subquery()
            )

            # Calculate lookback cutoff using the subquery
            lookback_cutoff = sa.func.now() - sa.func.make_interval(0, 0, 0, lookback_days_subquery)

            return sa.and_(
                KernelRow.scaling_group == scaling_group,
                KernelRow.starts_at.isnot(None),  # Must have started
                sa.or_(
                    # Running kernels (not yet terminated)
                    KernelRow.terminated_at.is_(None),
                    # Terminated kernels with unobserved period, within lookback
                    sa.and_(
                        KernelRow.terminated_at
                        > sa.func.coalesce(
                            KernelRow.last_observed_at,
                            KernelRow.starts_at,
                        ),
                        KernelRow.terminated_at >= lookback_cutoff,
                    ),
                ),
            )

        return inner


class KernelOrders:
    """Query orders for kernels."""

    @staticmethod
    def cluster_idx(ascending: bool = True) -> QueryOrder:
        if ascending:
            return KernelRow.cluster_idx.asc()
        return KernelRow.cluster_idx.desc()

    @staticmethod
    def created_at(ascending: bool = True) -> QueryOrder:
        if ascending:
            return KernelRow.created_at.asc()
        return KernelRow.created_at.desc()

    @staticmethod
    def terminated_at(ascending: bool = True) -> QueryOrder:
        if ascending:
            return KernelRow.terminated_at.asc()
        return KernelRow.terminated_at.desc()

    @staticmethod
    def status(ascending: bool = True) -> QueryOrder:
        if ascending:
            return KernelRow.status.asc()
        return KernelRow.status.desc()

    @staticmethod
    def cluster_mode(ascending: bool = True) -> QueryOrder:
        if ascending:
            return KernelRow.cluster_mode.asc()
        return KernelRow.cluster_mode.desc()

    @staticmethod
    def cluster_hostname(ascending: bool = True) -> QueryOrder:
        if ascending:
            return KernelRow.cluster_hostname.asc()
        return KernelRow.cluster_hostname.desc()


class UserConditions:
    """Query conditions for users."""

    @staticmethod
    def by_uuids(user_uuids: Collection[UUID]) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return UserRow.uuid.in_(user_uuids)

        return inner


class ImageConditions:
    """Query conditions for images."""

    @staticmethod
    def by_identifiers(identifiers: Collection[tuple[str, str]]) -> QueryCondition:
        """Filter images by list of (canonical, architecture) tuples.

        Args:
            identifiers: Collection of (canonical, architecture) tuples

        Returns:
            QueryCondition that matches any of the identifiers
        """

        def inner() -> sa.sql.expression.ColumnElement[bool]:
            if not identifiers:
                return sa.literal(False)

            conditions = [
                sa.and_(
                    ImageRow.name == canonical,
                    ImageRow.architecture == architecture,
                )
                for canonical, architecture in identifiers
            ]
            return sa.or_(*conditions)

        return inner



"""Email notification channel using SMTP."""

from __future__ import annotations

import asyncio
import logging
import smtplib
from email.mime.text import MIMEText
from functools import partial

from ai.backend.common.data.notification.types import EmailSpec
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.errors.notification import NotificationProcessingFailure
from ai.backend.manager.notification.channels.base import AbstractNotificationChannel
from ai.backend.manager.notification.types import NotificationMessage, SendResult

log = BraceStyleAdapter(logging.getLogger(__spec__.name))

_DEFAULT_NOTIFICATION_SUBJECT = "Notification from Backend.AI"


class EmailChannel(AbstractNotificationChannel):
    """Email notification channel using SMTP."""

    _spec: EmailSpec

    def __init__(self, email_spec: EmailSpec) -> None:
        """Initialize email channel with configuration."""
        self._spec = email_spec

    async def send(self, message: NotificationMessage) -> SendResult:
        """
        Send notification via email using SMTP.

        Args:
            message: Notification message to send

        Returns:
            SendResult indicating success

        Raises:
            NotificationProcessingFailure: If email delivery fails
        """
        # Determine subject (priority: spec template > first line of message)
        if self._spec.message.subject_template:
            subject = self._spec.message.subject_template
        else:
            # Fallback: use first line of message as subject
            lines = message.message.split("\n", 1)
            subject = lines[0] if lines else _DEFAULT_NOTIFICATION_SUBJECT

        loop = asyncio.get_running_loop()
        await loop.run_in_executor(
            None,
            partial(self._send_email, subject, message.message),
        )

        log.info(
            "Email notification sent successfully to {} recipients",
            len(self._spec.message.to_emails),
        )

        return SendResult(
            message=f"Email sent successfully to {len(self._spec.message.to_emails)} recipients"
        )

    def _send_email(self, subject: str, body: str) -> None:
        """
        Send email synchronously via SMTP.

        This method runs in a thread pool executor.
        """
        smtp = self._spec.smtp
        msg = self._spec.message

        email_msg = MIMEText(body, "plain", "utf-8")
        email_msg["Subject"] = subject
        email_msg["From"] = msg.from_email
        email_msg["To"] = ",".join(msg.to_emails)

        auth = self._spec.auth
        try:
            with smtplib.SMTP(smtp.host, smtp.port, timeout=smtp.timeout) as server:
                if smtp.use_tls:
                    server.starttls()
                if auth and auth.username and auth.password:
                    server.login(auth.username, auth.password)
                server.send_message(
                    email_msg,
                    from_addr=msg.from_email,
                    to_addrs=msg.to_emails,
                )
        except smtplib.SMTPConnectError as e:
            log.error(
                "Failed to connect to SMTP server {}:{}: {}",
                smtp.host,
                smtp.port,
                str(e),
            )
            raise NotificationProcessingFailure(f"SMTP connection failed: {e!s}") from e
        except smtplib.SMTPAuthenticationError as e:
            log.error(
                "SMTP authentication failed for user {}: {}",
                auth.username if auth else "unknown",
                str(e),
            )
            raise NotificationProcessingFailure(f"SMTP authentication failed: {e!s}") from e
        except smtplib.SMTPException as e:
            log.error("SMTP error while sending email: {}", str(e))
            raise NotificationProcessingFailure(f"Email delivery failed: {e!s}") from e
        except Exception as e:
            log.error("Unexpected error sending email: {}", str(e))
            raise NotificationProcessingFailure(f"Email delivery failed: {e!s}") from e



"""CreatorSpec implementations for domain repository."""

from __future__ import annotations

from dataclasses import dataclass
from typing import override

from ai.backend.common.types import ResourceSlot, VFolderHostPermissionMap
from ai.backend.manager.models.domain import DomainRow
from ai.backend.manager.repositories.base import CreatorSpec


@dataclass
class DomainCreatorSpec(CreatorSpec[DomainRow]):
    """CreatorSpec for domain creation."""

    name: str
    description: str | None = None
    is_active: bool | None = None
    total_resource_slots: ResourceSlot | None = None
    allowed_vfolder_hosts: dict[str, list[str]] | None = None
    allowed_docker_registries: list[str] | None = None
    integration_id: str | None = None
    dotfiles: bytes | None = None

    @override
    def build_row(self) -> DomainRow:
        return DomainRow(
            name=self.name,
            description=self.description,
            is_active=self.is_active if self.is_active is not None else True,
            total_resource_slots=self.total_resource_slots
            if self.total_resource_slots
            else ResourceSlot(),
            allowed_vfolder_hosts=self.allowed_vfolder_hosts
            if self.allowed_vfolder_hosts
            else VFolderHostPermissionMap(),
            allowed_docker_registries=self.allowed_docker_registries
            if self.allowed_docker_registries
            else [],
            integration_id=self.integration_id,
            dotfiles=self.dotfiles if self.dotfiles else b"\x90",
        )



from __future__ import annotations

from collections import defaultdict
from collections.abc import Sequence
from decimal import Decimal
from typing import TYPE_CHECKING, override
from uuid import UUID

from ai.backend.manager.data.fair_share import ProjectUserIds, UserFairShareFactors
from ai.backend.manager.sokovan.data import SessionWorkload, SystemSnapshot

from .sequencer import WorkloadSequencer

if TYPE_CHECKING:
    from ai.backend.manager.repositories.fair_share import FairShareRepository

# Default factor for users without computed factors (lowest priority)
# Factor 0 means no entitlement (maximum usage)
DEFAULT_FACTOR = Decimal("0")


class FairShareSequencer(WorkloadSequencer):
    """
    A scheduling sequencer that implements Fair Share sequencing.

    This sequencer orders workloads based on fair share factors fetched from the repository
    using a 3-way JOIN across domain, project, and user fair share tables.

    Sorting priority (descending - higher factor = higher priority):
    1. Domain fair share factor
    2. Project fair share factor
    3. User fair share factor

    Users with higher factors (lower historical usage) get scheduled first.
    If factors are not available for a user, they are placed at lowest priority.
    """

    _repository: FairShareRepository

    def __init__(self, repository: FairShareRepository) -> None:
        """Initialize the FairShareSequencer with the repository."""
        self._repository = repository

    @property
    @override
    def name(self) -> str:
        """Return the sequencer name for predicates."""
        return "FairShareSequencer"

    @override
    def success_message(self) -> str:
        """Return a message describing successful sequencing."""
        return "Sessions sequenced using Fair Share algorithm"

    @override
    async def sequence(
        self,
        resource_group: str,
        system_snapshot: SystemSnapshot,
        workloads: Sequence[SessionWorkload],
    ) -> Sequence[SessionWorkload]:
        """
        Sequence the workloads based on fair share factors.

        Workloads are sorted by (domain_factor, project_factor, user_factor) in descending order.
        Higher factor = higher priority (users with lower historical usage get scheduled first).

        :param resource_group: The resource group (scaling group) name.
        :param system_snapshot: The current system snapshot containing resource state.
        :param workloads: A sequence of SessionWorkload objects to sequence.
        :return: A sequence of SessionWorkload objects ordered by fair share factors.
        """
        if not workloads:
            return []

        # Load fair share factors from repository using 3-way JOIN
        user_factors = await self._load_factors(resource_group, workloads)

        # Sort by factors in descending order (higher factor = higher priority)
        # If a user doesn't have recorded factors, use default (lowest priority)
        return sorted(
            workloads,
            key=lambda w: self._get_sort_key(w.user_uuid, user_factors),
        )

    async def _load_factors(
        self,
        resource_group: str,
        workloads: Sequence[SessionWorkload],
    ) -> dict[UUID, UserFairShareFactors]:
        """Load fair share factors from the repository for the given workloads."""
        # Group user_ids by project_id
        project_users: dict[UUID, set[UUID]] = defaultdict(set)
        for w in workloads:
            project_users[w.group_id].add(w.user_uuid)

        # Build ProjectUserIds list
        project_user_ids = [
            ProjectUserIds(project_id=project_id, user_ids=frozenset(user_ids))
            for project_id, user_ids in project_users.items()
        ]

        # Fetch factors from repository using 3-way JOIN
        return await self._repository.get_user_fair_share_factors_batch(
            resource_group, project_user_ids
        )

    def _get_sort_key(
        self,
        user_uuid: UUID,
        user_factors: dict[UUID, UserFairShareFactors],
    ) -> tuple[Decimal, Decimal, Decimal]:
        """Get the sort key for a workload based on its user's factors.

        Returns negated factors for descending sort (higher factor = higher priority).
        """
        factors = user_factors.get(user_uuid)
        if factors is None:
            # No factors recorded - lowest priority
            return (-DEFAULT_FACTOR, -DEFAULT_FACTOR, -DEFAULT_FACTOR)
        return factors.sort_key()



from __future__ import annotations

from collections.abc import Sequence

from ai.backend.manager.data.domain.types import DomainData
from ai.backend.manager.repositories.base import BatchQuerier, NoPagination
from ai.backend.manager.repositories.domain.options import DomainConditions
from ai.backend.manager.services.domain.actions.search_domains import SearchDomainsAction
from ai.backend.manager.services.domain.processors import DomainProcessors


async def load_domains_by_names(
    processor: DomainProcessors,
    domain_names: Sequence[str],
) -> list[DomainData | None]:
    """Batch load domains by their names.

    Args:
        processor: The domain processor.
        domain_names: Sequence of domain names to load.

    Returns:
        List of DomainData (or None if not found) in the same order as domain_names.
    """
    if not domain_names:
        return []

    querier = BatchQuerier(
        pagination=NoPagination(),
        conditions=[DomainConditions.by_names(domain_names)],
    )

    action_result = await processor.search_domains.wait_for_complete(
        SearchDomainsAction(querier=querier)
    )

    domain_map = {domain.name: domain for domain in action_result.items}
    return [domain_map.get(name) for name in domain_names]



from __future__ import annotations

import enum
import uuid
from dataclasses import dataclass
from datetime import datetime
from typing import Any, override

from ai.backend.common.dto.manager.field import (
    VFolderOperationStatusField,
    VFolderOwnershipTypeField,
    VFolderPermissionField,
)
from ai.backend.common.types import CIStrEnum, QuotaScopeID, VFolderID, VFolderUsageMode
from ai.backend.manager.data.permission.types import OperationType
from ai.backend.manager.errors.resource import DataTransformationFailed


class VFolderOwnershipType(CIStrEnum):
    """
    Ownership type of virtual folder.
    """

    USER = "user"
    GROUP = "group"

    def to_field(self) -> VFolderOwnershipTypeField:
        return VFolderOwnershipTypeField(self)


class VFolderMountPermission(enum.StrEnum):
    # TODO: Replace this class with VFolderRBACPermission
    # Or rename this class to VFolderMountPermission
    """
    Permissions for a virtual folder given to a specific access key.
    RW_DELETE includes READ_WRITE and READ_WRITE includes READ_ONLY.
    """

    READ_ONLY = "ro"
    READ_WRITE = "rw"
    RW_DELETE = "wd"
    OWNER_PERM = "wd"  # resolved as RW_DELETE

    def to_field(self) -> VFolderPermissionField:
        return VFolderPermissionField(self)

    @override
    @classmethod
    def _missing_(cls, value: Any) -> VFolderMountPermission | None:
        if not isinstance(value, str):
            raise DataTransformationFailed(
                f"VFolderMountPermission value must be a string, got {type(value).__name__}"
            )
        match value.upper():
            case "RO" | "READ_ONLY":
                return cls.READ_ONLY
            case "RW" | "READ_WRITE":
                return cls.READ_WRITE
            case "RW_DELETE":
                return cls.RW_DELETE
            case "WD" | "OWNER_PERM":
                return cls.OWNER_PERM
        return None

    def to_rbac_operation(self) -> set[OperationType]:
        match self:
            case VFolderMountPermission.READ_ONLY:
                return {OperationType.READ}
            case VFolderMountPermission.READ_WRITE:
                return {OperationType.READ, OperationType.UPDATE, OperationType.SOFT_DELETE}
            case VFolderMountPermission.RW_DELETE | VFolderMountPermission.OWNER_PERM:
                return {OperationType.READ, OperationType.UPDATE, OperationType.SOFT_DELETE}


class VFolderInvitationState(enum.StrEnum):
    """
    Virtual Folder invitation state.
    """

    PENDING = "pending"
    CANCELED = "canceled"  # canceled by inviter
    ACCEPTED = "accepted"
    REJECTED = "rejected"  # rejected by invitee


class VFolderOperationStatus(enum.StrEnum):
    """
    Introduce virtual folder current status for storage-proxy operations.
    """

    READY = "ready"
    PERFORMING = "performing"
    CLONING = "cloning"
    MOUNTED = "mounted"
    ERROR = "error"

    DELETE_PENDING = "delete-pending"  # vfolder is in trash bin
    DELETE_ONGOING = "delete-ongoing"  # vfolder is being deleted in storage
    DELETE_COMPLETE = "delete-complete"  # vfolder is deleted permanently, only DB row remains
    DELETE_ERROR = "delete-error"

    @override
    @classmethod
    def _missing_(cls, value: Any) -> VFolderOperationStatus | None:
        if not isinstance(value, str):
            raise DataTransformationFailed(
                f"VFolderOperationStatus value must be a string, got {type(value).__name__}"
            )
        match value.upper():
            case "READY":
                return cls.READY
            case "PERFORMING":
                return cls.PERFORMING
            case "CLONING":
                return cls.CLONING
            case "MOUNTED":
                return cls.MOUNTED
            case "ERROR":
                return cls.ERROR
            case "DELETE_PENDING" | "DELETE-PENDING":
                return cls.DELETE_PENDING
            case "DELETE_ONGOING" | "DELETE-ONGOING":
                return cls.DELETE_ONGOING
            case "DELETE_COMPLETE" | "DELETE-COMPLETE":
                return cls.DELETE_COMPLETE
            case "DELETE_ERROR" | "DELETE-ERROR":
                return cls.DELETE_ERROR
        return None

    def is_deletable(self, force: bool = False) -> bool:
        if force:
            return self in {
                VFolderOperationStatus.READY,
                VFolderOperationStatus.DELETE_PENDING,
                VFolderOperationStatus.DELETE_ONGOING,
                VFolderOperationStatus.DELETE_ERROR,
            }
        return self == VFolderOperationStatus.DELETE_PENDING

    def to_field(self) -> VFolderOperationStatusField:
        return VFolderOperationStatusField(self)


@dataclass
class VFolderData:
    """
    Complete VFolder data representing all VFolder properties.
    Used by repository layer for returning full VFolder information.
    """

    id: uuid.UUID
    name: str
    host: str
    domain_name: str
    quota_scope_id: QuotaScopeID | None
    usage_mode: VFolderUsageMode
    permission: VFolderMountPermission | None
    max_files: int
    max_size: int | None
    num_files: int
    cur_size: int
    created_at: datetime
    last_used: datetime | None
    creator: str | None
    unmanaged_path: str | None
    ownership_type: VFolderOwnershipType
    user: uuid.UUID | None
    group: uuid.UUID | None
    cloneable: bool
    status: VFolderOperationStatus


@dataclass
class VFolderPermissionData:
    """
    VFolder permission data representing user-specific permissions on a VFolder.
    """

    id: uuid.UUID
    vfolder: uuid.UUID
    user: uuid.UUID
    permission: VFolderMountPermission


@dataclass
class VFolderInvitationData:
    """
    VFolder invitation data representing invitations to share VFolders.
    """

    id: uuid.UUID
    vfolder: uuid.UUID
    inviter: str  # email
    invitee: str  # email
    permission: VFolderMountPermission
    created_at: datetime
    modified_at: datetime | None


@dataclass
class VFolderCreateParams:
    """
    Parameters needed to create a new VFolder.
    """

    id: uuid.UUID
    name: str
    domain_name: str
    quota_scope_id: str
    usage_mode: VFolderUsageMode
    permission: VFolderMountPermission
    host: str
    creator: str
    ownership_type: VFolderOwnershipType
    user: uuid.UUID | None
    group: uuid.UUID | None
    unmanaged_path: str | None
    cloneable: bool
    status: VFolderOperationStatus


@dataclass
class VFolderAccessInfo:
    """
    Information about VFolder access for query results.
    """

    vfolder_data: VFolderData
    is_owner: bool
    effective_permission: VFolderMountPermission | None


@dataclass
class VFolderListResult:
    """
    Result of VFolder list operations with pagination support.
    """

    vfolders: list[VFolderAccessInfo]
    total_count: int | None = None


@dataclass
class VFolderDeleteParams:
    vfolder_id: VFolderID
    host: str
    unmanaged_path: str | None = None


class DeleteStatus(enum.StrEnum):
    DELETE_ONGOING = "delete_ongoing"
    ALREADY_DELETED = "already_deleted"
    ERROR = "error"


@dataclass
class VFolderDeleteResult:
    """
    Result of a VFolder delete operation.
    """

    vfolder_id: VFolderID
    status: DeleteStatus


@dataclass
class VFolderLocation:
    """
    Minimal VFolder location information for storage access.
    Contains only the essential fields needed to locate and access a vfolder in storage.
    """

    id: uuid.UUID
    quota_scope_id: QuotaScopeID | None
    host: str
    ownership_type: VFolderOwnershipType



"""Actions for searching projects."""

from __future__ import annotations

from dataclasses import dataclass
from typing import override
from uuid import UUID

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.data.group.types import GroupData
from ai.backend.manager.repositories.base.querier import BatchQuerier
from ai.backend.manager.repositories.group.types import (
    DomainProjectSearchScope,
    UserProjectSearchScope,
)
from ai.backend.manager.services.group.actions.base import GroupAction


@dataclass
class SearchProjectsAction(GroupAction):
    """Search all projects (admin scope - no scope filter)."""

    querier: BatchQuerier

    @override
    def entity_id(self) -> str | None:
        return None

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.SEARCH


@dataclass
class SearchProjectsByDomainAction(GroupAction):
    """Search projects within a domain."""

    scope: DomainProjectSearchScope
    querier: BatchQuerier

    @override
    def entity_id(self) -> str | None:
        return None

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.SEARCH


@dataclass
class SearchProjectsByUserAction(GroupAction):
    """Search projects a user is member of."""

    scope: UserProjectSearchScope
    querier: BatchQuerier

    @override
    def entity_id(self) -> str | None:
        return None

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.SEARCH


@dataclass
class GetProjectAction(GroupAction):
    """Get a single project by UUID."""

    project_id: UUID

    @override
    def entity_id(self) -> str | None:
        return str(self.project_id)

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.GET


# Result types


@dataclass
class SearchProjectsActionResult(BaseActionResult):
    """Result from searching projects."""

    items: list[GroupData]
    total_count: int
    has_next_page: bool
    has_previous_page: bool

    @override
    def entity_id(self) -> str | None:
        return None


@dataclass
class GetProjectActionResult(BaseActionResult):
    """Result from getting a single project."""

    data: GroupData

    @override
    def entity_id(self) -> str | None:
        return str(self.data.id)



"""add_vfolder_host_permission

Revision ID: 5bce905c21e5
Revises: 360af8f33d4e
Create Date: 2022-10-13 18:35:21.955941

"""

from typing import Any

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql as pgsql
from sqlalchemy.sql.expression import bindparam

from ai.backend.common.types import VFolderHostPermission
from ai.backend.manager.models.base import GUID, metadata

# revision identifiers, used by Alembic.
revision = "5bce905c21e5"
down_revision = "360af8f33d4e"
branch_labels = None
depends_on = None


ALL_HOST_PERMISSIONS = [perm.value for perm in VFolderHostPermission]


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    domains = sa.Table(
        "domains",
        metadata,
        sa.Column("name", sa.String(length=64), primary_key=True),
        sa.Column("allowed_vfolder_hosts", pgsql.ARRAY(sa.String)),
        extend_existing=True,
    )
    groups = sa.Table(
        "groups",
        metadata,
        sa.Column("id", GUID, primary_key=True),
        sa.Column("allowed_vfolder_hosts", pgsql.ARRAY(sa.String)),
        extend_existing=True,
    )
    keypair_resource_policies = sa.Table(
        "keypair_resource_policies",
        metadata,
        sa.Column("name", sa.String(length=256), primary_key=True),
        sa.Column("allowed_vfolder_hosts", pgsql.ARRAY(sa.String)),
        extend_existing=True,
    )
    domain_query = sa.select(domains.c.name, domains.c.allowed_vfolder_hosts)
    group_query = sa.select(groups.c.id, groups.c.allowed_vfolder_hosts)
    keypair_resource_policies_query = sa.select(
        keypair_resource_policies.c.name,
        keypair_resource_policies.c.allowed_vfolder_hosts,
    )

    connection = op.get_bind()
    bind_param_pk = "bind_param_pk"

    def get_pk_vfolder_host_maps(
        query: sa.sql.Select[Any], pk_name: str
    ) -> list[dict[str, object]]:
        map_list = []
        rows = connection.execute(query).fetchall()
        for row in rows:
            hosts: list[str] = row.allowed_vfolder_hosts
            all_perms: dict[str, list[str]] = dict.fromkeys(hosts, ALL_HOST_PERMISSIONS)
            map_list.append({
                bind_param_pk: getattr(row, pk_name),
                "allowed_vfolder_hosts": all_perms,
            })
        return map_list

    domain_list = get_pk_vfolder_host_maps(domain_query, "name")
    group_list = get_pk_vfolder_host_maps(group_query, "id")
    resource_policy_list = get_pk_vfolder_host_maps(keypair_resource_policies_query, "name")

    op.drop_column("domains", "allowed_vfolder_hosts")
    op.drop_column("groups", "allowed_vfolder_hosts")
    op.drop_column("keypair_resource_policies", "allowed_vfolder_hosts")

    op.add_column(
        "domains",
        sa.Column(
            "allowed_vfolder_hosts",
            pgsql.JSONB(),
            nullable=False,
            server_default=sa.text("'{}'::jsonb"),
        ),
    )
    op.add_column(
        "groups",
        sa.Column(
            "allowed_vfolder_hosts",
            pgsql.JSONB(),
            nullable=False,
            server_default=sa.text("'{}'::jsonb"),
        ),
    )
    op.add_column(
        "keypair_resource_policies",
        sa.Column(
            "allowed_vfolder_hosts",
            pgsql.JSONB(),
            nullable=False,
            server_default=sa.text("'{}'::jsonb"),
        ),
    )

    domains = sa.Table(
        "domains",
        metadata,
        sa.Column("name", sa.String(length=64), primary_key=True),
        sa.Column("allowed_vfolder_hosts", pgsql.JSONB),
        extend_existing=True,
    )
    groups = sa.Table(
        "groups",
        metadata,
        sa.Column("id", GUID, primary_key=True),
        sa.Column("allowed_vfolder_hosts", pgsql.JSONB),
        extend_existing=True,
    )
    keypair_resource_policies = sa.Table(
        "keypair_resource_policies",
        metadata,
        sa.Column("name", sa.String(length=256), primary_key=True),
        sa.Column("allowed_vfolder_hosts", pgsql.JSONB),
        extend_existing=True,
    )

    def update_vfolder_host_col(
        table: sa.Table, cond: sa.ColumnElement[Any], row_list: list[dict[str, object]]
    ) -> None:
        query = (
            sa.update(table)
            .values({"allowed_vfolder_hosts": bindparam("allowed_vfolder_hosts")})
            .where(cond)
        )
        connection.execute(query, row_list)

    update_vfolder_host_col(domains, domains.c.name == bindparam(bind_param_pk), domain_list)
    update_vfolder_host_col(groups, groups.c.id == bindparam(bind_param_pk), group_list)
    update_vfolder_host_col(
        keypair_resource_policies,
        keypair_resource_policies.c.name == bindparam(bind_param_pk),
        resource_policy_list,
    )

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    domains = sa.Table(
        "domains",
        metadata,
        sa.Column("name", sa.String(length=64), primary_key=True),
        sa.Column("allowed_vfolder_hosts", pgsql.JSONB),
        extend_existing=True,
    )
    groups = sa.Table(
        "groups",
        metadata,
        sa.Column("id", GUID, primary_key=True),
        sa.Column("allowed_vfolder_hosts", pgsql.JSONB),
        extend_existing=True,
    )
    keypair_resource_policies = sa.Table(
        "keypair_resource_policies",
        metadata,
        sa.Column("name", sa.String(length=256), primary_key=True),
        sa.Column("allowed_vfolder_hosts", pgsql.JSONB),
        extend_existing=True,
    )
    connection = op.get_bind()
    bind_param_pk = "bind_param_pk"

    def get_pk_vfolder_host_maps(
        query: sa.sql.Select[Any], pk_name: str
    ) -> list[dict[str, object]]:
        map_list = []
        rows = connection.execute(query).fetchall()
        for row in rows:
            hosts: dict[str, list[str]] = row.allowed_vfolder_hosts
            map_list.append({
                bind_param_pk: getattr(row, pk_name),
                "allowed_vfolder_hosts": list(hosts.keys()),
            })
        return map_list

    domain_query = sa.select(domains.c.name, domains.c.allowed_vfolder_hosts)
    group_query = sa.select(groups.c.id, groups.c.allowed_vfolder_hosts)
    keypair_resource_policies_query = sa.select(
        keypair_resource_policies.c.name,
        keypair_resource_policies.c.allowed_vfolder_hosts,
    )

    domain_list = get_pk_vfolder_host_maps(domain_query, "name")
    group_list = get_pk_vfolder_host_maps(group_query, "id")
    resource_policy_list = get_pk_vfolder_host_maps(keypair_resource_policies_query, "name")

    op.drop_column("domains", "allowed_vfolder_hosts")
    op.drop_column("groups", "allowed_vfolder_hosts")
    op.drop_column("keypair_resource_policies", "allowed_vfolder_hosts")

    op.add_column(
        "domains",
        sa.Column(
            "allowed_vfolder_hosts", pgsql.ARRAY(sa.String), nullable=False, server_default="{}"
        ),
    )
    op.add_column(
        "groups",
        sa.Column(
            "allowed_vfolder_hosts", pgsql.ARRAY(sa.String), nullable=False, server_default="{}"
        ),
    )
    op.add_column(
        "keypair_resource_policies",
        sa.Column(
            "allowed_vfolder_hosts", pgsql.ARRAY(sa.String), nullable=False, server_default="{}"
        ),
    )

    domains = sa.Table(
        "domains",
        metadata,
        sa.Column("name", sa.String(length=64), primary_key=True),
        sa.Column("allowed_vfolder_hosts", pgsql.ARRAY(sa.String)),
        extend_existing=True,
    )
    groups = sa.Table(
        "groups",
        metadata,
        sa.Column("id", GUID, primary_key=True),
        sa.Column("allowed_vfolder_hosts", pgsql.ARRAY(sa.String)),
        extend_existing=True,
    )
    keypair_resource_policies = sa.Table(
        "keypair_resource_policies",
        metadata,
        sa.Column("name", sa.String(length=256), primary_key=True),
        sa.Column("allowed_vfolder_hosts", pgsql.ARRAY(sa.String)),
        extend_existing=True,
    )

    def update_vfolder_host_col(
        table: sa.Table, cond: sa.ColumnElement[Any], row_list: list[dict[str, object]]
    ) -> None:
        query = (
            sa.update(table)
            .values({"allowed_vfolder_hosts": bindparam("allowed_vfolder_hosts")})
            .where(cond)
        )
        connection.execute(query, row_list)

    update_vfolder_host_col(domains, domains.c.name == bindparam(bind_param_pk), domain_list)
    update_vfolder_host_col(groups, groups.c.id == bindparam(bind_param_pk), group_list)
    update_vfolder_host_col(
        keypair_resource_policies,
        keypair_resource_policies.c.name == bindparam(bind_param_pk),
        resource_policy_list,
    )

    # ### end Alembic commands ###



"""add-internal-data-to-kernels

Revision ID: 202b6dcbc159
Revises: 3f1dafab60b2
Create Date: 2019-10-01 16:13:20.935285

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "202b6dcbc159"
down_revision = "3f1dafab60b2"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "kernels",
        sa.Column("internal_data", postgresql.JSONB(astext_type=sa.Text()), nullable=True),
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("kernels", "internal_data")
    # ### end Alembic commands ###



"""Deployment service for managing model deployments."""

import logging
from datetime import UTC, datetime

from ai.backend.common.data.endpoint.types import EndpointLifecycle
from ai.backend.common.data.model_deployment.types import (
    ActivenessStatus,
    DeploymentStrategy,
    LivenessStatus,
    ModelDeploymentStatus,
    ReadinessStatus,
)
from ai.backend.common.data.permission.types import EntityType, ScopeType
from ai.backend.common.types import (
    ResourceSlot,
)
from ai.backend.logging.utils import BraceStyleAdapter
from ai.backend.manager.data.deployment.types import (
    ClusterConfigData,
    DeploymentInfo,
    ExtraVFolderMountData,
    ModelDeploymentAccessTokenData,
    ModelDeploymentData,
    ModelDeploymentMetadataInfo,
    ModelMountConfigData,
    ModelReplicaData,
    ModelRevisionData,
    ModelRuntimeConfigData,
    ReplicaStateData,
    ResourceConfigData,
    RouteInfo,
)
from ai.backend.manager.data.permission.id import ScopeId
from ai.backend.manager.errors.service import RoutingNotFound
from ai.backend.manager.models.endpoint import EndpointRow, EndpointTokenRow
from ai.backend.manager.repositories.base import Creator
from ai.backend.manager.repositories.base.rbac.entity_creator import RBACEntityCreator
from ai.backend.manager.repositories.deployment import DeploymentRepository
from ai.backend.manager.repositories.deployment.creators import (
    DeploymentCreatorSpec,
    DeploymentExecutionFields,
    DeploymentMetadataFields,
    DeploymentMountFields,
    DeploymentNetworkFields,
    DeploymentReplicaFields,
    DeploymentResourceFields,
    EndpointTokenCreatorSpec,
    ModelRevisionFields,
)
from ai.backend.manager.services.deployment.actions.access_token.create_access_token import (
    CreateAccessTokenAction,
    CreateAccessTokenActionResult,
)
from ai.backend.manager.services.deployment.actions.access_token.search_access_tokens import (
    SearchAccessTokensAction,
    SearchAccessTokensActionResult,
)
from ai.backend.manager.services.deployment.actions.auto_scaling_rule.create_auto_scaling_rule import (
    CreateAutoScalingRuleAction,
    CreateAutoScalingRuleActionResult,
)
from ai.backend.manager.services.deployment.actions.auto_scaling_rule.delete_auto_scaling_rule import (
    DeleteAutoScalingRuleAction,
    DeleteAutoScalingRuleActionResult,
)
from ai.backend.manager.services.deployment.actions.auto_scaling_rule.search_auto_scaling_rules import (
    SearchAutoScalingRulesAction,
    SearchAutoScalingRulesActionResult,
)
from ai.backend.manager.services.deployment.actions.auto_scaling_rule.update_auto_scaling_rule import (
    UpdateAutoScalingRuleAction,
    UpdateAutoScalingRuleActionResult,
)
from ai.backend.manager.services.deployment.actions.create_deployment import (
    CreateDeploymentAction,
    CreateDeploymentActionResult,
)
from ai.backend.manager.services.deployment.actions.create_legacy_deployment import (
    CreateLegacyDeploymentAction,
    CreateLegacyDeploymentActionResult,
)
from ai.backend.manager.services.deployment.actions.deployment_policy import (
    GetDeploymentPolicyAction,
    GetDeploymentPolicyActionResult,
)
from ai.backend.manager.services.deployment.actions.destroy_deployment import (
    DestroyDeploymentAction,
    DestroyDeploymentActionResult,
)
from ai.backend.manager.services.deployment.actions.get_deployment_by_id import (
    GetDeploymentByIdAction,
    GetDeploymentByIdActionResult,
)
from ai.backend.manager.services.deployment.actions.get_replica_by_id import (
    GetReplicaByIdAction,
    GetReplicaByIdActionResult,
)
from ai.backend.manager.services.deployment.actions.model_revision.add_model_revision import (
    AddModelRevisionAction,
    AddModelRevisionActionResult,
)
from ai.backend.manager.services.deployment.actions.model_revision.create_model_revision import (
    CreateModelRevisionAction,
    CreateModelRevisionActionResult,
)
from ai.backend.manager.services.deployment.actions.model_revision.get_revision_by_id import (
    GetRevisionByIdAction,
    GetRevisionByIdActionResult,
)
from ai.backend.manager.services.deployment.actions.model_revision.search_revisions import (
    SearchRevisionsAction,
    SearchRevisionsActionResult,
)
from ai.backend.manager.services.deployment.actions.revision_operations import (
    ActivateRevisionAction,
    ActivateRevisionActionResult,
)
from ai.backend.manager.services.deployment.actions.route import (
    SearchRoutesAction,
    SearchRoutesActionResult,
    UpdateRouteTrafficStatusAction,
    UpdateRouteTrafficStatusActionResult,
)
from ai.backend.manager.services.deployment.actions.search_deployments import (
    SearchDeploymentsAction,
    SearchDeploymentsActionResult,
)
from ai.backend.manager.services.deployment.actions.search_replicas import (
    SearchReplicasAction,
    SearchReplicasActionResult,
)
from ai.backend.manager.services.deployment.actions.sync_replicas import (
    SyncReplicaAction,
    SyncReplicaActionResult,
)
from ai.backend.manager.services.deployment.actions.update_deployment import (
    UpdateDeploymentAction,
    UpdateDeploymentActionResult,
)
from ai.backend.manager.sokovan.deployment import DeploymentController
from ai.backend.manager.sokovan.deployment.types import DeploymentLifecycleType

log = BraceStyleAdapter(logging.getLogger(__name__))


def _map_lifecycle_to_status(lifecycle: EndpointLifecycle) -> ModelDeploymentStatus:
    """Map EndpointLifecycle to ModelDeploymentStatus."""
    match lifecycle:
        case EndpointLifecycle.PENDING:
            return ModelDeploymentStatus.PENDING
        case EndpointLifecycle.CREATED | EndpointLifecycle.READY:
            return ModelDeploymentStatus.READY
        case EndpointLifecycle.SCALING:
            return ModelDeploymentStatus.SCALING
        case EndpointLifecycle.DEPLOYING:
            return ModelDeploymentStatus.DEPLOYING
        case EndpointLifecycle.DESTROYING:
            return ModelDeploymentStatus.STOPPING
        case EndpointLifecycle.DESTROYED:
            return ModelDeploymentStatus.STOPPED


def _convert_deployment_info_to_data(info: DeploymentInfo) -> ModelDeploymentData:
    """Convert DeploymentInfo to ModelDeploymentData.

    Note: Some fields are set to defaults as DeploymentInfo doesn't have all the data.
    """
    # Map revision if available
    revision: ModelRevisionData | None = None
    if info.model_revisions:
        rev = info.model_revisions[0]
        revision = ModelRevisionData(
            id=info.current_revision_id or info.id,
            name=rev.image_identifier.canonical,
            cluster_config=ClusterConfigData(
                mode=rev.resource_spec.cluster_mode,
                size=rev.resource_spec.cluster_size,
            ),
            resource_config=ResourceConfigData(
                resource_group_name=info.metadata.resource_group,
                resource_slot=ResourceSlot.from_json(rev.resource_spec.resource_slots),
            ),
            model_mount_config=ModelMountConfigData(
                vfolder_id=rev.mounts.model_vfolder_id,
                mount_destination=rev.mounts.model_mount_destination,
                definition_path=rev.mounts.model_definition_path or "",
            ),
            model_runtime_config=ModelRuntimeConfigData(
                runtime_variant=rev.execution.runtime_variant,
                inference_runtime_config=rev.execution.inference_runtime_config or {},
            ),
            extra_vfolder_mounts=[
                ExtraVFolderMountData(
                    vfolder_id=m.vfid.folder_id,
                    mount_destination=m.kernel_path.as_posix(),
                )
                for m in rev.mounts.extra_mounts
            ],
            image_id=info.current_revision_id
            or info.id,  # Placeholder: actual image_id not in ImageIdentifier
            created_at=info.metadata.created_at or datetime.now(UTC),
        )

    desired_count = info.replica_spec.desired_replica_count
    if desired_count is None:
        desired_count = info.replica_spec.replica_count

    return ModelDeploymentData(
        id=info.id,
        metadata=ModelDeploymentMetadataInfo(
            name=info.metadata.name,
            status=_map_lifecycle_to_status(info.state.lifecycle),
            tags=[info.metadata.tag] if info.metadata.tag else [],
            project_id=info.metadata.project,
            domain_name=info.metadata.domain,
            created_at=info.metadata.created_at or datetime.now(UTC),
            updated_at=info.metadata.created_at or datetime.now(UTC),
        ),
        network_access=info.network,
        revision_history_ids=[info.current_revision_id] if info.current_revision_id else [],
        revision=revision,
        scaling_rule_ids=[],  # Not available in DeploymentInfo
        replica_state=ReplicaStateData(
            desired_replica_count=desired_count,
            replica_ids=[],  # Not available in DeploymentInfo
        ),
        default_deployment_strategy=DeploymentStrategy.ROLLING,
        created_user_id=info.metadata.created_user,
    )


def _convert_route_info_to_replica_data(route: RouteInfo) -> ModelReplicaData:
    """Convert RouteInfo to ModelReplicaData.

    Note: Some fields are set to defaults as RouteInfo doesn't have all the data.
    """
    return ModelReplicaData(
        id=route.route_id,
        revision_id=route.revision_id
        or route.endpoint_id,  # Fallback to endpoint_id if no revision
        session_id=route.session_id or route.route_id,
        readiness_status=ReadinessStatus.HEALTHY,  # Derived from route status
        liveness_status=LivenessStatus.HEALTHY,  # Default
        activeness_status=ActivenessStatus.ACTIVE
        if route.traffic_ratio > 0
        else ActivenessStatus.INACTIVE,
        weight=int(route.traffic_ratio * 100),  # Convert ratio to weight
        detail=route.error_data,
        created_at=route.created_at or datetime.now(tz=UTC),
    )


class DeploymentService:
    """Service for managing deployments."""

    _deployment_controller: DeploymentController
    _deployment_repository: DeploymentRepository

    def __init__(
        self,
        deployment_controller: DeploymentController,
        deployment_repository: DeploymentRepository,
    ) -> None:
        """Initialize deployment service with controller and repository."""
        self._deployment_controller = deployment_controller
        self._deployment_repository = deployment_repository

    # ========== Deployment CRUD ==========

    async def create_deployment(
        self, action: CreateDeploymentAction
    ) -> CreateDeploymentActionResult:
        """Create a new deployment.

        Args:
            action: Create deployment action containing the creator specification

        Returns:
            CreateDeploymentActionResult: Result containing the created deployment data
        """
        log.info("Creating deployment with name: {}", action.creator.metadata.name)

        # Build CreatorSpec from action data
        metadata = action.creator.metadata
        revision = action.creator.model_revision
        mounts = revision.mounts

        creator_spec = DeploymentCreatorSpec(
            metadata=DeploymentMetadataFields(
                name=metadata.name,
                domain=metadata.domain,
                project_id=metadata.project,
                resource_group=metadata.resource_group,
                created_user_id=metadata.created_user,
                session_owner_id=metadata.session_owner,
                revision_history_limit=metadata.revision_history_limit,
                tag=metadata.tag,
            ),
            replica=DeploymentReplicaFields(
                replica_count=action.creator.replica_spec.replica_count,
                desired_replica_count=action.creator.replica_spec.desired_replica_count,
            ),
            network=DeploymentNetworkFields(
                open_to_public=action.creator.network.open_to_public,
                url=action.creator.network.url,
            ),
            revision=ModelRevisionFields(
                image_id=revision.image_id,
                resource=DeploymentResourceFields(
                    cluster_mode=revision.resource_spec.cluster_mode,
                    cluster_size=revision.resource_spec.cluster_size,
                    resource_slots=ResourceSlot(revision.resource_spec.resource_slots),
                    resource_opts=revision.resource_spec.resource_opts,
                ),
                mounts=DeploymentMountFields(
                    model_vfolder_id=mounts.model_vfolder_id,
                    model_mount_destination=mounts.model_mount_destination,
                    model_definition_path=mounts.model_definition_path,
                    extra_mounts=(),  # TODO: Convert MountInfo to VFolderMount
                ),
                execution=DeploymentExecutionFields(
                    runtime_variant=revision.execution.runtime_variant,
                    startup_command=revision.execution.startup_command,
                    bootstrap_script=revision.execution.bootstrap_script,
                    environ=revision.execution.environ,
                    callback_url=revision.execution.callback_url,
                ),
            ),
        )
        creator: RBACEntityCreator[EndpointRow] = RBACEntityCreator(
            spec=creator_spec,
            scope_ref=ScopeId(scope_type=ScopeType.USER, scope_id=str(metadata.created_user)),
            additional_scope_refs=[],
            entity_type=EntityType.MODEL_DEPLOYMENT,
        )

        # Create endpoint via repository
        deployment_info = await self._deployment_repository.create_endpoint(
            creator, action.creator.policy
        )

        # Mark lifecycle needed to start provisioning
        await self._deployment_controller.mark_lifecycle_needed(
            DeploymentLifecycleType.CHECK_PENDING
        )

        return CreateDeploymentActionResult(data=_convert_deployment_info_to_data(deployment_info))

    async def create_legacy_deployment(
        self, action: CreateLegacyDeploymentAction
    ) -> CreateLegacyDeploymentActionResult:
        """Create a new legacy deployment(Model Serving).

        Args:
            action: Create legacy deployment action containing the creator specification

        Returns:
            CreateLegacyDeploymentActionResult: Result containing the created deployment info
        """
        log.info("Creating deployment with name: {}", action.draft.name)
        deployment_info = await self._deployment_controller.create_deployment(action.draft)
        await self._deployment_controller.mark_lifecycle_needed(
            DeploymentLifecycleType.CHECK_PENDING
        )
        return CreateLegacyDeploymentActionResult(data=deployment_info)

    async def update_deployment(
        self, action: UpdateDeploymentAction
    ) -> UpdateDeploymentActionResult:
        """Update an existing deployment.

        Args:
            action: Update deployment action containing the updater

        Returns:
            UpdateDeploymentActionResult: Result containing the updated deployment data
        """
        log.info("Updating deployment with ID: {}", action.updater.pk_value)

        # Update endpoint and get updated deployment info in one call
        deployment_info = await self._deployment_repository.update_endpoint(action.updater)

        # Mark lifecycle needed for potential replica adjustments
        await self._deployment_controller.mark_lifecycle_needed(
            DeploymentLifecycleType.CHECK_REPLICA
        )

        return UpdateDeploymentActionResult(data=_convert_deployment_info_to_data(deployment_info))

    async def destroy_deployment(
        self, action: DestroyDeploymentAction
    ) -> DestroyDeploymentActionResult:
        """Destroy an existing deployment.

        Args:
            action: Destroy deployment action containing the endpoint ID

        Returns:
            DestroyDeploymentActionResult: Result indicating success or failure
        """
        log.info("Destroying deployment with ID: {}", action.endpoint_id)
        success = await self._deployment_controller.destroy_deployment(action.endpoint_id)
        await self._deployment_controller.mark_lifecycle_needed(DeploymentLifecycleType.DESTROYING)
        return DestroyDeploymentActionResult(success=success)

    async def search_deployments(
        self, action: SearchDeploymentsAction
    ) -> SearchDeploymentsActionResult:
        """Search deployments with filtering and pagination.

        Args:
            action: Action containing BatchQuerier for filtering and pagination

        Returns:
            SearchDeploymentsActionResult: Result containing list of deployments and pagination info
        """
        result = await self._deployment_repository.search_endpoints(action.querier)
        deployments = [_convert_deployment_info_to_data(info) for info in result.items]
        return SearchDeploymentsActionResult(
            data=deployments,
            total_count=result.total_count,
            has_next_page=result.has_next_page,
            has_previous_page=result.has_previous_page,
        )

    async def get_deployment_by_id(
        self, action: GetDeploymentByIdAction
    ) -> GetDeploymentByIdActionResult:
        """Get a deployment by ID.

        Args:
            action: Action containing the deployment ID

        Returns:
            GetDeploymentByIdActionResult: Result containing the deployment data

        Raises:
            EndpointNotFound: If the deployment does not exist
        """
        deployment_info = await self._deployment_repository.get_endpoint_info(action.deployment_id)
        return GetDeploymentByIdActionResult(data=_convert_deployment_info_to_data(deployment_info))

    async def get_deployment_policy(
        self, action: GetDeploymentPolicyAction
    ) -> GetDeploymentPolicyActionResult:
        """Get the deployment policy for an endpoint.

        Args:
            action: Action containing the endpoint ID

        Returns:
            GetDeploymentPolicyActionResult: Result containing the policy data

        Raises:
            DeploymentPolicyNotFound: If the policy does not exist
        """
        data = await self._deployment_controller.get_deployment_policy(action.endpoint_id)
        return GetDeploymentPolicyActionResult(data=data)

    # ========== Revision Operations ==========

    async def add_model_revision(
        self, action: AddModelRevisionAction
    ) -> AddModelRevisionActionResult:
        # TODO: Implement full revision creation logic
        # This requires integration with the controller's revision generator system:
        # 1. Get default architecture from scaling group
        # 2. Use revision generator to resolve image and build ModelRevisionSpec
        # 3. Get latest revision number via get_latest_revision_number()
        # 4. Build DeploymentRevisionCreatorSpec and create revision
        raise NotImplementedError(
            "add_model_revision requires controller's revision generator for image resolution. "
            "Use create_legacy_deployment for deployment creation with revision."
        )

    async def create_model_revision(
        self, action: CreateModelRevisionAction
    ) -> CreateModelRevisionActionResult:
        # TODO: Implement full revision creation logic
        # Note: CreateModelRevisionAction is missing deployment_id field
        # This requires integration with the controller's revision generator system
        raise NotImplementedError(
            "create_model_revision requires controller's revision generator for image resolution "
            "and is missing deployment_id in action definition."
        )

    async def get_revision_by_id(
        self, action: GetRevisionByIdAction
    ) -> GetRevisionByIdActionResult:
        revision = await self._deployment_repository.get_revision(action.revision_id)
        return GetRevisionByIdActionResult(data=revision)

    async def search_revisions(self, action: SearchRevisionsAction) -> SearchRevisionsActionResult:
        """Search revisions with filtering and pagination.

        Args:
            action: Action containing BatchQuerier for filtering and pagination

        Returns:
            SearchRevisionsActionResult: Result containing list of revisions and pagination info
        """
        result = await self._deployment_repository.search_revisions(action.querier)
        return SearchRevisionsActionResult(
            data=result.items,
            total_count=result.total_count,
            has_next_page=result.has_next_page,
            has_previous_page=result.has_previous_page,
        )

    async def activate_revision(
        self, action: ActivateRevisionAction
    ) -> ActivateRevisionActionResult:
        """Activate a specific revision to be the current revision.

        Args:
            action: Action containing deployment and revision IDs

        Returns:
            ActivateRevisionActionResult: Result containing the updated deployment
        """
        # 1. Validate revision exists (raises exception if not found)
        _revision = await self._deployment_repository.get_revision(action.revision_id)

        # 2. Update endpoint.current_revision and get previous revision
        previous_revision_id = await self._deployment_repository.update_current_revision(
            action.deployment_id, action.revision_id
        )

        # 3. Trigger lifecycle check to update routes with new revision
        await self._deployment_controller.mark_lifecycle_needed(
            DeploymentLifecycleType.CHECK_REPLICA
        )

        log.info(
            "Activated revision {} for deployment {} (previous: {})",
            action.revision_id,
            action.deployment_id,
            previous_revision_id,
        )

        # 4. Get updated deployment info
        deployment_info = await self._deployment_repository.get_endpoint_info(action.deployment_id)

        return ActivateRevisionActionResult(
            deployment=_convert_deployment_info_to_data(deployment_info),
            previous_revision_id=previous_revision_id,
            activated_revision_id=action.revision_id,
        )

    # ========== Route Operations ==========

    async def sync_replicas(self, action: SyncReplicaAction) -> SyncReplicaActionResult:
        """Sync replicas for a deployment.

        This triggers a lifecycle check to reconcile the actual replica count
        with the desired replica count based on the current revision.

        Args:
            action: Action containing the deployment ID

        Returns:
            SyncReplicaActionResult: Result indicating success
        """
        # Trigger lifecycle check to sync replicas
        await self._deployment_controller.mark_lifecycle_needed(
            DeploymentLifecycleType.CHECK_REPLICA
        )

        log.info("Triggered replica sync for deployment {}", action.deployment_id)

        return SyncReplicaActionResult(success=True)

    async def search_routes(self, action: SearchRoutesAction) -> SearchRoutesActionResult:
        """Search routes with filtering and pagination.

        Args:
            action: Action containing BatchQuerier for filtering and pagination

        Returns:
            SearchRoutesActionResult: Result containing list of routes and pagination info
        """
        result = await self._deployment_repository.search_routes(action.querier)
        return SearchRoutesActionResult(
            routes=result.items,
            total_count=result.total_count,
            has_next_page=result.has_next_page,
            has_previous_page=result.has_previous_page,
        )

    async def update_route_traffic_status(
        self, action: UpdateRouteTrafficStatusAction
    ) -> UpdateRouteTrafficStatusActionResult:
        """Update route traffic status.

        Args:
            action: Action containing route ID and new traffic status

        Returns:
            UpdateRouteTrafficStatusActionResult: Result containing updated route

        Raises:
            RouteNotFound: If the route does not exist
        """
        route = await self._deployment_controller.update_route_traffic_status(
            route_id=action.route_id,
            traffic_status=action.traffic_status,
        )
        if route is None:
            raise RoutingNotFound
        return UpdateRouteTrafficStatusActionResult(route=route)

    # ========== Auto-scaling Rules ==========

    async def create_auto_scaling_rule(
        self, action: CreateAutoScalingRuleAction
    ) -> CreateAutoScalingRuleActionResult:
        """Create a new auto-scaling rule for a deployment.

        Args:
            action: Action containing the rule creator specification

        Returns:
            CreateAutoScalingRuleActionResult: Result containing the created rule data
        """
        data = await self._deployment_repository.create_model_deployment_autoscaling_rule(
            action.creator
        )
        return CreateAutoScalingRuleActionResult(data=data)

    async def update_auto_scaling_rule(
        self, action: UpdateAutoScalingRuleAction
    ) -> UpdateAutoScalingRuleActionResult:
        """Update an existing auto-scaling rule.

        Args:
            action: Action containing the rule ID and modifier

        Returns:
            UpdateAutoScalingRuleActionResult: Result containing the updated rule data
        """
        data = await self._deployment_repository.update_model_deployment_autoscaling_rule(
            action.auto_scaling_rule_id,
            action.modifier,
        )
        return UpdateAutoScalingRuleActionResult(data=data)

    async def delete_auto_scaling_rule(
        self, action: DeleteAutoScalingRuleAction
    ) -> DeleteAutoScalingRuleActionResult:
        """Delete an auto-scaling rule.

        Args:
            action: Action containing the rule ID to delete

        Returns:
            DeleteAutoScalingRuleActionResult: Result indicating success
        """
        success = await self._deployment_repository.delete_autoscaling_rule(
            action.auto_scaling_rule_id
        )
        return DeleteAutoScalingRuleActionResult(success=success)

    # ========== Access Token ==========

    async def create_access_token(
        self, action: CreateAccessTokenAction
    ) -> CreateAccessTokenActionResult:
        """Create a new access token for a model deployment.

        Args:
            action: CreateAccessTokenAction containing the creator spec.

        Returns:
            CreateAccessTokenActionResult with the created token data.
        """
        # Get endpoint info to retrieve domain, project, session_owner
        endpoint_info = await self._deployment_repository.get_endpoint_info(
            action.creator.model_deployment_id
        )

        # Create the Creator with EndpointTokenCreatorSpec
        spec = EndpointTokenCreatorSpec(
            endpoint_id=action.creator.model_deployment_id,
            domain=endpoint_info.metadata.domain,
            project_id=endpoint_info.metadata.project,
            session_owner_id=endpoint_info.metadata.session_owner,
        )
        creator: Creator[EndpointTokenRow] = Creator(spec=spec)

        # Create the token via repository
        token_row = await self._deployment_repository.create_access_token(creator)

        # Convert to ModelDeploymentAccessTokenData
        # Note: valid_until is returned as provided but not persisted in DB
        data = ModelDeploymentAccessTokenData(
            id=token_row.id,
            token=token_row.token,
            valid_until=action.creator.valid_until,
            created_at=token_row.created_at or datetime.now(UTC),
        )
        return CreateAccessTokenActionResult(data=data)

    # ========== Replica Operations ==========

    async def get_replica_by_id(self, action: GetReplicaByIdAction) -> GetReplicaByIdActionResult:
        """Get a replica by ID."""
        route = await self._deployment_repository.get_route(action.replica_id)
        if route is None:
            return GetReplicaByIdActionResult(data=None)
        return GetReplicaByIdActionResult(data=_convert_route_info_to_replica_data(route))

    # ========== Search Operations ==========

    async def search_replicas(self, action: SearchReplicasAction) -> SearchReplicasActionResult:
        """Search replicas with pagination, ordering, and filtering."""
        result = await self._deployment_repository.search_routes(action.querier)
        replicas = [_convert_route_info_to_replica_data(route) for route in result.items]
        return SearchReplicasActionResult(
            data=replicas,
            total_count=result.total_count,
            has_next_page=result.has_next_page,
            has_previous_page=result.has_previous_page,
        )

    async def search_access_tokens(
        self, action: SearchAccessTokensAction
    ) -> SearchAccessTokensActionResult:
        """Search access tokens with pagination and ordering."""
        result = await self._deployment_repository.search_access_tokens(action.querier)
        return SearchAccessTokensActionResult(
            data=result.items,
            total_count=result.total_count,
            has_next_page=result.has_next_page,
            has_previous_page=result.has_previous_page,
        )

    async def search_auto_scaling_rules(
        self, action: SearchAutoScalingRulesAction
    ) -> SearchAutoScalingRulesActionResult:
        """Search auto-scaling rules with pagination and ordering."""
        result = await self._deployment_repository.search_auto_scaling_rules(action.querier)
        return SearchAutoScalingRulesActionResult(
            data=result.items,
            total_count=result.total_count,
            has_next_page=result.has_next_page,
            has_previous_page=result.has_previous_page,
        )



"""Deployment-specific recorder for deployment coordinator."""

from __future__ import annotations

from uuid import UUID

from ai.backend.manager.sokovan.recorder.recorder import TransitionRecorder

# Deployment-specific type alias
DeploymentTransitionRecorder = TransitionRecorder[UUID]

__all__ = [
    "DeploymentTransitionRecorder",
]



from typing import override

from ai.backend.manager.actions.monitors.monitor import ActionMonitor
from ai.backend.manager.actions.processor import ActionProcessor
from ai.backend.manager.actions.types import AbstractProcessorPackage, ActionSpec
from ai.backend.manager.services.storage_namespace.actions.get_all import (
    GetAllNamespacesAction,
    GetAllNamespacesActionResult,
)
from ai.backend.manager.services.storage_namespace.actions.get_multi import (
    GetNamespacesAction,
    GetNamespacesActionResult,
)
from ai.backend.manager.services.storage_namespace.actions.register import (
    RegisterNamespaceAction,
    RegisterNamespaceActionResult,
)
from ai.backend.manager.services.storage_namespace.actions.search import (
    SearchStorageNamespacesAction,
    SearchStorageNamespacesActionResult,
)
from ai.backend.manager.services.storage_namespace.actions.unregister import (
    UnregisterNamespaceAction,
    UnregisterNamespaceActionResult,
)
from ai.backend.manager.services.storage_namespace.service import StorageNamespaceService


class StorageNamespaceProcessors(AbstractProcessorPackage):
    register: ActionProcessor[RegisterNamespaceAction, RegisterNamespaceActionResult]
    unregister: ActionProcessor[UnregisterNamespaceAction, UnregisterNamespaceActionResult]
    get_namespaces: ActionProcessor[GetNamespacesAction, GetNamespacesActionResult]
    get_all_namespaces: ActionProcessor[GetAllNamespacesAction, GetAllNamespacesActionResult]
    search_storage_namespaces: ActionProcessor[
        SearchStorageNamespacesAction, SearchStorageNamespacesActionResult
    ]

    def __init__(
        self, service: StorageNamespaceService, action_monitors: list[ActionMonitor]
    ) -> None:
        self.register = ActionProcessor(service.register, action_monitors)
        self.unregister = ActionProcessor(service.unregister, action_monitors)
        self.get_namespaces = ActionProcessor(service.get_namespaces, action_monitors)
        self.get_all_namespaces = ActionProcessor(service.get_all_namespaces, action_monitors)
        self.search_storage_namespaces = ActionProcessor(service.search, action_monitors)

    @override
    def supported_actions(self) -> list[ActionSpec]:
        return [
            RegisterNamespaceAction.spec(),
            UnregisterNamespaceAction.spec(),
            GetNamespacesAction.spec(),
            GetAllNamespacesAction.spec(),
            SearchStorageNamespacesAction.spec(),
        ]



"""add domain, group, user fields to kernels

Revision ID: c5e4e764f9e3
Revises: 6f1c1b83870a
Create Date: 2019-05-28 10:22:56.904061

"""

import textwrap

import sqlalchemy as sa
from alembic import op
from sqlalchemy.sql import text

from ai.backend.manager.models.base import GUID, ForeignKeyIDColumn, IDColumn, convention

# revision identifiers, used by Alembic.
revision = "c5e4e764f9e3"
down_revision = "6f1c1b83870a"
branch_labels = None
depends_on = None


def upgrade() -> None:
    metadata = sa.MetaData(naming_convention=convention)
    # partial tables for data migration
    _groups = sa.Table(
        "groups",
        metadata,
        IDColumn("id"),
        sa.Column("name", sa.String(length=64), nullable=False),
        sa.Column("description", sa.String(length=512)),
        sa.Column("is_active", sa.Boolean, default=True),
        sa.Column(
            "domain_name",
            sa.String(length=64),
            sa.ForeignKey("domains.name", onupdate="CASCADE", ondelete="CASCADE"),
            nullable=False,
            index=True,
        ),
    )
    users = sa.Table(
        "users",
        metadata,
        IDColumn("uuid"),
    )
    kernels = sa.Table(
        "kernels",
        metadata,
        IDColumn(),
        sa.Column("access_key", sa.String(length=20), sa.ForeignKey("keypairs.access_key")),
    )
    keypairs = sa.Table(
        "keypairs",
        metadata,
        sa.Column("user_id", sa.String(length=256), index=True),
        sa.Column("access_key", sa.String(length=20), primary_key=True),
        ForeignKeyIDColumn("user", "users.uuid", nullable=False),
    )

    op.add_column("kernels", sa.Column("domain_name", sa.String(length=64), nullable=True))
    op.add_column("kernels", sa.Column("group_id", GUID(), nullable=True))
    op.add_column("kernels", sa.Column("user_uuid", GUID(), nullable=True))
    op.create_foreign_key(
        op.f("fk_kernels_group_id_groups"), "kernels", "groups", ["group_id"], ["id"]
    )
    op.create_foreign_key(
        op.f("fk_kernels_user_uuid_users"), "kernels", "users", ["user_uuid"], ["uuid"]
    )
    op.create_foreign_key(
        op.f("fk_kernels_domain_name_domains"), "kernels", "domains", ["domain_name"], ["name"]
    )

    # Create default group in the default domain.
    # Assumption: "default" domain must exist
    connection = op.get_bind()
    insert_query = textwrap.dedent(
        """\
        INSERT INTO groups (name, description, is_active, domain_name)
        VALUES ('default', 'Default group', True, 'default')
        ON CONFLICT (name, domain_name) DO NOTHING
        RETURNING id;
    """
    )
    result = connection.execute(text(insert_query)).first()
    gid = result.id if result is not None else None
    if gid is None:  # group already exists
        select_query = textwrap.dedent(
            """\
            SELECT id FROM groups where name = 'default' and domain_name = 'default';
        """
        )
        result = connection.execute(text(select_query)).first()
        if result is not None:
            gid = result.id

    # Fill in kernels' domain_name, group_id, and user_uuid.
    select_kernels_query = sa.select(kernels.c.id, kernels.c.access_key).select_from(kernels)
    all_kernels = connection.execute(select_kernels_query).fetchall()
    for kernel in all_kernels:
        # Get kernel's keypair (access_key).
        select_keypair_query = (
            sa.select(keypairs.c.user)
            .select_from(keypairs)
            .where(keypairs.c.access_key == kernel.access_key)
        )
        kp = connection.execute(select_keypair_query).first()
        if kp is not None:
            # Update kernel information.
            update_query = f"""\
                UPDATE kernels SET domain_name = 'default', group_id = '{gid}', user_uuid = '{kp.user}'
                WHERE id = '{kernel.id}';
            """
            connection.execute(text(update_query))

    # Associate every users with the default group.
    # NOTE: this operation is not undoable unless you drop groups table.
    select_users_query = sa.select(users.c.uuid).select_from(users)
    all_users = connection.execute(select_users_query).fetchall()
    for user in all_users:
        insert_assoc_query = f"""\
            INSERT INTO association_groups_users (group_id, user_id)
            VALUES ('{gid}', '{user.uuid}')
            ON CONFLICT (group_id, user_id) DO NOTHING;
        """
        connection.execute(text(insert_assoc_query))

    # Make kernel's new fields non-nullable.
    op.alter_column("kernels", column_name="domain_name", nullable=False)
    op.alter_column("kernels", column_name="group_id", nullable=False)
    op.alter_column("kernels", column_name="user_uuid", nullable=False)


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint(op.f("fk_kernels_domain_name_domains"), "kernels", type_="foreignkey")
    op.drop_constraint(op.f("fk_kernels_user_uuid_users"), "kernels", type_="foreignkey")
    op.drop_constraint(op.f("fk_kernels_group_id_groups"), "kernels", type_="foreignkey")
    op.drop_column("kernels", "user_uuid")
    op.drop_column("kernels", "group_id")
    op.drop_column("kernels", "domain_name")
    # ### end Alembic commands ###
